import inspect
import typing
from abc import ABC, abstractmethod

from telegrinder.bot.cute_types import BaseCute, MessageCute, UpdateCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.process import check_rule
from telegrinder.bot.rules.adapter import ABCAdapter, EventAdapter, RawUpdateAdapter
from telegrinder.tools.i18n.base import ABCTranslator
from telegrinder.tools.magic import cache_translation, get_cached_translation
from telegrinder.types.objects import Update as UpdateObject

T = typing.TypeVar("T", bound=BaseCute)

Message: typing.TypeAlias = MessageCute
Update: typing.TypeAlias = UpdateCute


def with_caching_translations(func):
    """Should be used as decorator for .translate method. Caches rule translations."""

    async def wrapper(self: "ABCRule[typing.Any]", translator: ABCTranslator):
        if translation := get_cached_translation(self, translator.locale):
            return translation
        translation = await func(self, translator)
        cache_translation(self, translator.locale, translation)
        return translation

    return wrapper


class ABCRule(ABC, typing.Generic[T]):
    adapter: ABCAdapter[UpdateObject, T] = RawUpdateAdapter()  # type: ignore
    requires: list["ABCRule[T]"] = []

    @abstractmethod
    async def check(self, event: T, ctx: Context) -> bool:
        pass

    def __init_subclass__(cls, requires: list["ABCRule[T]"] | None = None):
        """Merges requirements from inherited classes and rule-specific requirements."""

        requirements = []
        for base in inspect.getmro(cls):
            if issubclass(base, ABCRule) and base != cls:
                requirements.extend(base.requires or ())  # type: ignore

        requirements.extend(requires or ())
        cls.requires = list(dict.fromkeys(requirements))

    def __and__(self, other: "ABCRule[T]"):
        return AndRule(self, other)

    def __or__(self, other: "ABCRule[T]"):
        return OrRule(self, other)

    def __neg__(self) -> "ABCRule[T]":
        return NotRule(self)

    def __repr__(self) -> str:
        return "<rule: {!r}, adapter: {!r}>".format(
            self.__class__.__name__,
            self.adapter,
        )

    async def translate(self, translator: ABCTranslator) -> typing.Self:
        return self


class AndRule(ABCRule[T]):
    def __init__(self, *rules: ABCRule[T]):
        self.rules = rules

    async def check(self, event: Update, ctx: Context) -> bool:
        ctx_copy = ctx.copy()
        for rule in self.rules:
            if not await check_rule(event.ctx_api, rule, event, ctx_copy):
                return False
        ctx |= ctx_copy
        return True


class OrRule(ABCRule[T]):
    def __init__(self, *rules: ABCRule[T]):
        self.rules = rules

    async def check(self, event: Update, ctx: Context) -> bool:
        for rule in self.rules:
            ctx_copy = ctx.copy()
            if await check_rule(event.ctx_api, rule, event, ctx_copy):
                ctx |= ctx_copy
                return True
        return False


class NotRule(ABCRule[T]):
    def __init__(self, rule: ABCRule[T]):
        self.rule = rule

    async def check(self, event: Update, ctx: Context) -> bool:
        ctx_copy = ctx.copy()
        return not await check_rule(event.ctx_api, self.rule, event, ctx_copy)


__all__ = (
    "ABCRule",
    "AndRule",
    "NotRule",
    "OrRule",
    "with_caching_translations",
)



from .base import ABCI18n, ABCTranslator, I18nEnum
from .middleware import ABCTranslatorMiddleware
from .simple import SimpleI18n, SimpleTranslator

__all__ = (
    "ABCI18n",
    "ABCTranslator",
    "ABCTranslatorMiddleware",
    "I18nEnum",
    "SimpleI18n",
    "SimpleTranslator",
)



from telegrinder.bot.cute_types import InlineQueryCute
from telegrinder.bot.dispatch.return_manager import InlineQueryReturnManager

from .abc import BaseStateView


class InlineQueryView(BaseStateView[InlineQueryCute]):
    def __init__(self) -> None:
        self.auto_rules = []
        self.handlers = []
        self.middlewares = []
        self.return_manager = InlineQueryReturnManager()

    def get_state_key(self, event: InlineQueryCute) -> int | None:
        return event.from_user.id


__all__ = ("InlineQueryView",)



import typing

import msgspec
from fntypes.result import Error, Ok, Result

from telegrinder.api.response import APIResponse
from telegrinder.client import ABCClient, AiohttpClient
from telegrinder.model import DataConverter, decoder
from telegrinder.types.methods import APIMethods

from .abc import ABCAPI, APIError, Token


def compose_data(
    client: ABCClient,
    data: dict[str, typing.Any],
    files: dict[str, tuple[str, bytes]],
) -> typing.Any:
    converter = DataConverter(files=files.copy())
    return client.get_form(
        data={k: converter(v) for k, v in data.items()},
        files=converter.files,
    )


class API(ABCAPI, APIMethods):
    """Bot API with available API methods."""

    API_URL = "https://api.telegram.org/"

    def __init__(self, token: Token, *, http: ABCClient | None = None) -> None:
        self.token = token
        self.http = http or AiohttpClient()
        super().__init__(self)

    def __repr__(self) -> str:
        return "<{}: token={!r}, http={!r}>".format(
            self.__class__.__name__,
            self.token,
            self.http,
        )

    @property
    def id(self) -> int:
        return self.token.bot_id

    @property
    def request_url(self) -> str:
        return self.API_URL + f"bot{self.token}/"

    async def request(
        self,
        method: str,
        data: dict[str, typing.Any] | None = None,
        files: dict[str, tuple[str, bytes]] | None = None,
    ) -> Result[dict[str, typing.Any] | list[typing.Any] | bool, APIError]:
        response = await self.http.request_json(
            url=self.request_url + method,
            data=compose_data(self.http, data or {}, files or {}),
        )
        if response.get("ok"):
            assert "result" in response
            return Ok(response["result"])
        return Error(
            APIError(
                code=response.get("error_code", 400),
                error=response.get("description"),
            )
        )

    async def request_raw(
        self,
        method: str,
        data: dict[str, typing.Any] | None = None,
        files: dict[str, tuple[str, bytes]] | None = None,
    ) -> Result[msgspec.Raw, APIError]:
        response_bytes = await self.http.request_bytes(
            url=self.request_url + method,
            data=compose_data(self.http, data or {}, files or {}),
        )
        return decoder.decode(response_bytes, type=APIResponse).to_result()


__all__ = ("API",)



import typing
from abc import ABC, abstractmethod

from fntypes.result import Result

from telegrinder.api import ABCAPI
from telegrinder.bot.dispatch.context import Context

EventT = typing.TypeVar("EventT")
Handler = typing.Callable[typing.Concatenate[EventT, ...], typing.Awaitable[typing.Any]]


class ABCErrorHandler(ABC, typing.Generic[EventT]):
    @abstractmethod
    def __call__(
        self,
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> typing.Callable[[typing.Callable[..., typing.Any]], typing.Callable[..., typing.Any]]:
        """Decorator for registering callback as an error handler."""

    @abstractmethod
    async def run(
        self,
        handler: Handler[EventT],
        event: EventT,
        api: ABCAPI,
        ctx: Context,
    ) -> Result[typing.Any, typing.Any]:
        """Run error handler."""


__all__ = ("ABCErrorHandler",)



import typing

from telegrinder.model import get_params
from telegrinder.types import (
    ChatPermissions,
    InlineKeyboardMarkup,
    InlineQueryResultArticle,
    InlineQueryResultAudio,
    InlineQueryResultCachedAudio,
    InlineQueryResultCachedDocument,
    InlineQueryResultCachedGif,
    InlineQueryResultCachedMpeg4Gif,
    InlineQueryResultCachedPhoto,
    InlineQueryResultCachedSticker,
    InlineQueryResultCachedVideo,
    InlineQueryResultCachedVoice,
    InlineQueryResultContact,
    InlineQueryResultDocument,
    InlineQueryResultGame,
    InlineQueryResultGif,
    InlineQueryResultLocation,
    InlineQueryResultMpeg4Gif,
    InlineQueryResultPhoto,
    InlineQueryResultVenue,
    InlineQueryResultVideo,
    InlineQueryResultVoice,
    InputContactMessageContent,
    InputFile,
    InputInvoiceMessageContent,
    InputLocationMessageContent,
    InputMediaAnimation,
    InputMediaAudio,
    InputMediaDocument,
    InputMediaPhoto,
    InputMediaVideo,
    InputTextMessageContent,
    InputVenueMessageContent,
    LabeledPrice,
    LinkPreviewOptions,
    MessageEntity,
    ReactionEmoji,
    ReactionType,
    ReactionTypeEmoji,
    ReplyParameters,
)

InputMedia: typing.TypeAlias = typing.Union[
    InputMediaAnimation,
    InputMediaAudio,
    InputMediaDocument,
    InputMediaPhoto,
    InputMediaVideo,
]
InputMessageContent: typing.TypeAlias = typing.Union[
    InputTextMessageContent,
    InputLocationMessageContent,
    InputVenueMessageContent,
    InputContactMessageContent,
    InputInvoiceMessageContent,
]

INPUT_MEDIA_TYPES: typing.Final[dict[str, type[InputMedia]]] = {
    "animation": InputMediaAnimation,
    "audio": InputMediaAudio,
    "document": InputMediaDocument,
    "photo": InputMediaPhoto,
    "video": InputMediaVideo,
}


def compose_reactions(
    reactions: (str | ReactionEmoji | ReactionType | list[str | ReactionEmoji | ReactionType]),
    /,
) -> list[ReactionType]:
    if not isinstance(reactions, list):
        reactions = [reactions]
    return [
        (
            ReactionTypeEmoji("emoji", emoji)
            if isinstance(emoji, ReactionEmoji)
            else (
                ReactionTypeEmoji("emoji", ReactionEmoji(emoji))
                if isinstance(emoji, str)
                else emoji
            )
        )
        for emoji in ([reactions] if isinstance(reactions, str) else reactions)
    ]


def compose_reply_params(
    message_id: int | None,
    chat_id: int | str | None,
    *,
    allow_sending_without_reply: bool | None = None,
    quote: str | None = None,
    quote_parse_mode: str | None = None,
    quote_entities: list[MessageEntity] | None = None,
    quote_position: int | None = None,
    **other: typing.Any,
) -> ReplyParameters:
    return ReplyParameters(**get_params(locals()))


def compose_link_preview_options(
    *,
    is_disabled: bool | None = None,
    url: str | None = None,
    prefer_small_media: bool | None = None,
    prefer_large_media: bool | None = None,
    show_above_text: bool | None = None,
    **other: typing.Any,
) -> LinkPreviewOptions:
    return LinkPreviewOptions(**get_params(locals()))


def compose_chat_permissions(
    *,
    can_send_messages: bool | None = None,
    can_send_audios: bool | None = None,
    can_send_documents: bool | None = None,
    can_send_photos: bool | None = None,
    can_send_videos: bool | None = None,
    can_send_video_notes: bool | None = None,
    can_send_voice_notes: bool | None = None,
    can_send_polls: bool | None = None,
    can_send_other_messages: bool | None = None,
    can_add_web_page_previews: bool | None = None,
    can_change_info: bool | None = None,
    can_invite_users: bool | None = None,
    can_pin_messages: bool | None = None,
    can_manage_topics: bool | None = None,
) -> ChatPermissions:
    return ChatPermissions(**get_params(locals()))


def input_media(
    type: typing.Literal["animation", "audio", "document", "photo", "video"],
    media: str | InputFile,
    *,
    caption: str | None = None,
    parse_mode: str | None = None,
    caption_entities: list[MessageEntity] | None = None,
    **other: typing.Any,
) -> InputMedia:
    return INPUT_MEDIA_TYPES[type](**get_params(locals()))


def input_contact_message_content(
    phone_number: str,
    first_name: str,
    *,
    last_name: str | None = None,
    vcard: str | None = None,
) -> InputContactMessageContent:
    return InputContactMessageContent(**get_params(locals()))


def input_invoice_message_content(
    title: str,
    description: str,
    payload: str,
    provider_token: str,
    currency: str,
    prices: list[LabeledPrice],
    *,
    max_tip_amount: int | None = None,
    suggested_tip_amounts: list[int] | None = None,
    provider_data: str | None = None,
    photo_url: str | None = None,
    photo_size: int | None = None,
    photo_width: int | None = None,
    photo_height: int | None = None,
    need_name: bool | None = None,
    need_phone_number: bool | None = None,
    need_email: bool | None = None,
    need_shipping_address: bool | None = None,
    send_phone_number_to_provider: bool | None = None,
) -> InputInvoiceMessageContent:
    return InputInvoiceMessageContent(**get_params(locals()))


def input_location_message_content(
    latitude: float,
    longitude: float,
    *,
    horizontal_accuracy: float | None = None,
    live_period: int | None = None,
    heading: int | None = None,
    proximity_alert_radius: int | None = None,
) -> InputLocationMessageContent:
    return InputLocationMessageContent(**get_params(locals()))


def input_text_message_content(
    message_text: str,
    *,
    parse_mode: str | None = None,
    entities: list[MessageEntity] | None = None,
    disable_web_page_preview: bool | None = None,
) -> InputTextMessageContent:
    return InputTextMessageContent(**get_params(locals()))


def input_venue_message_content(
    latitude: float,
    longitude: float,
    title: str,
    address: str,
    *,
    foursquare_id: str | None = None,
    foursquare_type: str | None = None,
    google_place_id: str | None = None,
    google_place_type: str | None = None,
) -> InputVenueMessageContent:
    return InputVenueMessageContent(**get_params(locals()))


def inline_query_article(
    id: str,
    title: str,
    input_message_content: InputMessageContent,
    *,
    reply_markup: InlineKeyboardMarkup | None = None,
    url: str | None = None,
    hide_url: bool | None = None,
    description: str | None = None,
    thumbnail_url: str | None = None,
    thumbnail_width: int | None = None,
    thumbnail_height: int | None = None,
) -> InlineQueryResultArticle:
    return InlineQueryResultArticle(type="article", **get_params(locals()))


def inline_query_audio(
    id: str,
    audio_url: str,
    *,
    title: str | None = None,
    caption: str | None = None,
    parse_mode: str | None = None,
    caption_entities: list[MessageEntity] | None = None,
    performer: str | None = None,
    audio_duration: int | None = None,
    reply_markup: InlineKeyboardMarkup | None = None,
    input_message_content: InputMessageContent | None = None,
) -> InlineQueryResultAudio:
    return InlineQueryResultAudio(type="audio", **get_params(locals()))


def inline_query_contact(
    id: str,
    phone_number: str,
    first_name: str,
    *,
    last_name: str | None = None,
    vcard: str | None = None,
    reply_markup: InlineKeyboardMarkup | None = None,
    input_message_content: InputMessageContent | None = None,
    thumbnail_url: str | None = None,
    thumbnail_width: int | None = None,
    thumbnail_height: int | None = None,
) -> InlineQueryResultContact:
    return InlineQueryResultContact(type="contact", **get_params(locals()))


def inline_query_document(
    id: str,
    title: str,
    document_url: str,
    mime_type: str,
    *,
    description: str | None = None,
    caption: str | None = None,
    parse_mode: str | None = None,
    caption_entities: list[MessageEntity] | None = None,
    reply_markup: InlineKeyboardMarkup | None = None,
    input_message_content: InputMessageContent | None = None,
    thumbnail_url: str | None = None,
    thumbnail_width: int | None = None,
    thumbnail_height: int | None = None,
) -> InlineQueryResultDocument:
    return InlineQueryResultDocument(type="document", **get_params(locals()))


def inline_query_gif(
    id: str,
    gif_url: str,
    *,
    gif_width: int | None = None,
    gif_height: int | None = None,
    gif_duration: int | None = None,
    title: str | None = None,
    caption: str | None = None,
    parse_mode: str | None = None,
    caption_entities: list[MessageEntity] | None = None,
    reply_markup: InlineKeyboardMarkup | None = None,
    input_message_content: InputMessageContent | None = None,
    thumbnail_url: str | None = None,
    thumbnail_mime_type: str | None = None,
) -> InlineQueryResultGif:
    return InlineQueryResultGif(type="gif", **get_params(locals()))


def inline_query_location(
    id: str,
    latitude: float,
    longitude: float,
    *,
    title: str | None = None,
    horizontal_accuracy: float | None = None,
    live_period: int | None = None,
    heading: int | None = None,
    proximity_alert_radius: int | None = None,
    reply_markup: InlineKeyboardMarkup | None = None,
    input_message_content: InputMessageContent | None = None,
    thumbnail_url: str | None = None,
    thumbnail_width: int | None = None,
    thumbnail_height: int | None = None,
) -> InlineQueryResultLocation:
    return InlineQueryResultLocation(type="location", **get_params(locals()))


def inline_query_venue(
    id: str,
    latitude: float,
    longitude: float,
    title: str,
    address: str,
    *,
    foursquare_id: str | None = None,
    foursquare_type: str | None = None,
    google_place_id: str | None = None,
    google_place_type: str | None = None,
    reply_markup: InlineKeyboardMarkup | None = None,
    input_message_content: InputMessageContent | None = None,
    thumbnail_url: str | None = None,
    thumbnail_width: int | None = None,
    thumbnail_height: int | None = None,
) -> InlineQueryResultVenue:
    return InlineQueryResultVenue(type="venue", **get_params(locals()))


def inline_query_video(
    id: str,
    video_url: str,
    mime_type: str,
    thumb_url: str,
    *,
    title: str | None = None,
    caption: str | None = None,
    parse_mode: str | None = None,
    caption_entities: list[MessageEntity] | None = None,
    video_width: int | None = None,
    video_height: int | None = None,
    video_duration: int | None = None,
    description: str | None = None,
    reply_markup: InlineKeyboardMarkup | None = None,
    input_message_content: InputMessageContent | None = None,
    thumbnail_url: str | None = None,
    thumbnail_width: int | None = None,
    thumbnail_height: int | None = None,
) -> InlineQueryResultVideo:
    return InlineQueryResultVideo(type="video", **get_params(locals()))


def inline_query_game(
    id: str,
    game_short_name: str,
    *,
    reply_markup: InlineKeyboardMarkup | None = None,
) -> InlineQueryResultGame:
    return InlineQueryResultGame(type="game", **get_params(locals()))


def inline_query_voice(
    id: str,
    voice_url: str,
    title: str,
    *,
    caption: str | None = None,
    parse_mode: str | None = None,
    caption_entities: list[MessageEntity] | None = None,
    reply_markup: InlineKeyboardMarkup | None = None,
    input_message_content: InputMessageContent | None = None,
    duration: int | None = None,
) -> InlineQueryResultVoice:
    return InlineQueryResultVoice(type="voice", **get_params(locals()))


def inline_query_photo(
    id: str,
    photo_url: str,
    thumb_url: str,
    *,
    photo_width: int | None = None,
    photo_height: int | None = None,
    title: str | None = None,
    description: str | None = None,
    caption: str | None = None,
    parse_mode: str | None = None,
    caption_entities: list[MessageEntity] | None = None,
    reply_markup: InlineKeyboardMarkup | None = None,
    input_message_content: InputMessageContent | None = None,
) -> InlineQueryResultPhoto:
    return InlineQueryResultPhoto(type="photo", **get_params(locals()))


def inline_query_mpeg4_gif(
    id: str,
    mpeg4_url: str,
    *,
    mpeg4_width: int | None = None,
    mpeg4_height: int | None = None,
    mpeg4_duration: int | None = None,
    thumb_url: str | None = None,
    reply_markup: InlineKeyboardMarkup | None = None,
    input_message_content: InputMessageContent | None = None,
) -> InlineQueryResultMpeg4Gif:
    return InlineQueryResultMpeg4Gif(type="mpeg4_gif", **get_params(locals()))


def inline_query_cached_sticker(
    id: str,
    sticker_file_id: str,
    *,
    reply_markup: InlineKeyboardMarkup | None = None,
) -> InlineQueryResultCachedSticker:
    return InlineQueryResultCachedSticker(type="sticker", **get_params(locals()))


def inline_query_cached_document(
    id: str,
    document_file_id: str,
    *,
    title: str | None = None,
    description: str | None = None,
    caption: str | None = None,
    parse_mode: str | None = None,
    caption_entities: list[MessageEntity] | None = None,
    reply_markup: InlineKeyboardMarkup | None = None,
    input_message_content: InputMessageContent | None = None,
) -> InlineQueryResultCachedDocument:
    return InlineQueryResultCachedDocument(type="document", **get_params(locals()))


def inline_query_cached_audio(
    id: str,
    audio_file_id: str,
    *,
    caption: str | None = None,
    parse_mode: str | None = None,
    caption_entities: list[MessageEntity] | None = None,
    reply_markup: InlineKeyboardMarkup | None = None,
    input_message_content: InputMessageContent | None = None,
) -> InlineQueryResultCachedAudio:
    return InlineQueryResultCachedAudio(type="audio", **get_params(locals()))


def inline_query_cached_video(
    id: str,
    video_file_id: str,
    *,
    title: str | None = None,
    description: str | None = None,
    caption: str | None = None,
    parse_mode: str | None = None,
    caption_entities: list[MessageEntity] | None = None,
    reply_markup: InlineKeyboardMarkup | None = None,
    input_message_content: InputMessageContent | None = None,
) -> InlineQueryResultCachedVideo:
    return InlineQueryResultCachedVideo(type="video", **get_params(locals()))


def inline_query_cached_gif(
    id: str,
    gif_file_id: str,
    *,
    title: str | None = None,
    caption: str | None = None,
    parse_mode: str | None = None,
    caption_entities: list[MessageEntity] | None = None,
    reply_markup: InlineKeyboardMarkup | None = None,
    input_message_content: InputMessageContent | None = None,
) -> InlineQueryResultCachedGif:
    return InlineQueryResultCachedGif(type="gif", **get_params(locals()))


def inline_query_cached_mpeg4_gif(
    id: str,
    mpeg4_file_id: str,
    *,
    title: str | None = None,
    caption: str | None = None,
    parse_mode: str | None = None,
    caption_entities: list[MessageEntity] | None = None,
    reply_markup: InlineKeyboardMarkup | None = None,
    input_message_content: InputMessageContent | None = None,
) -> InlineQueryResultCachedMpeg4Gif:
    return InlineQueryResultCachedMpeg4Gif(type="mpeg4_gif", **get_params(locals()))


def inline_query_cached_voice(
    id: str,
    voice_file_id: str,
    *,
    caption: str | None = None,
    parse_mode: str | None = None,
    caption_entities: list[MessageEntity] | None = None,
    reply_markup: InlineKeyboardMarkup | None = None,
    input_message_content: InputMessageContent | None = None,
) -> InlineQueryResultCachedVoice:
    return InlineQueryResultCachedVoice(type="voice", **get_params(locals()))


def inline_query_cached_photo(
    id: str,
    photo_file_id: str,
    *,
    title: str | None = None,
    description: str | None = None,
    caption: str | None = None,
    parse_mode: str | None = None,
    caption_entities: list[MessageEntity] | None = None,
    reply_markup: InlineKeyboardMarkup | None = None,
    input_message_content: InputMessageContent | None = None,
) -> InlineQueryResultCachedPhoto:
    return InlineQueryResultCachedPhoto(type="photo", **get_params(locals()))


__all__ = (
    "compose_chat_permissions",
    "compose_link_preview_options",
    "compose_reactions",
    "compose_reply_params",
    "inline_query_article",
    "inline_query_audio",
    "inline_query_cached_audio",
    "inline_query_cached_document",
    "inline_query_cached_gif",
    "inline_query_cached_mpeg4_gif",
    "inline_query_cached_photo",
    "inline_query_cached_sticker",
    "inline_query_cached_video",
    "inline_query_cached_voice",
    "inline_query_contact",
    "inline_query_document",
    "inline_query_game",
    "inline_query_gif",
    "inline_query_location",
    "inline_query_mpeg4_gif",
    "inline_query_photo",
    "inline_query_venue",
    "inline_query_video",
    "inline_query_voice",
    "input_contact_message_content",
    "input_invoice_message_content",
    "input_location_message_content",
    "input_media",
    "input_text_message_content",
    "input_venue_message_content",
)



import dataclasses
import typing

from telegrinder.bot.dispatch.context import Context

from .abc import ABCRule, T, Update, check_rule
from .func import FuncRule


@dataclasses.dataclass
class RuleEnumState:
    name: str
    rule: ABCRule
    cls: type["RuleEnum"]

    def __eq__(self, other: typing.Self) -> bool:
        return self.cls == other.cls and self.name == other.name


class RuleEnum(ABCRule[T]):
    __enum__: list[RuleEnumState]

    def __init_subclass__(cls, *args, **kwargs):
        new_attributes = set(cls.__dict__) - set(RuleEnum.__dict__) - {"__enum__", "__init__"}
        enum_lst: list[RuleEnumState] = []

        self = cls.__new__(cls)
        self.__init__()

        for attribute_name in new_attributes:
            rules = getattr(cls, attribute_name)
            attribute = RuleEnumState(attribute_name, rules, cls)

            setattr(
                self,
                attribute.name,
                self & FuncRule(lambda _, ctx: self.must_be_state(ctx, attribute)),
            )
            enum_lst.append(attribute)

        setattr(cls, "__enum__", enum_lst)

    @classmethod
    def save_state(cls, ctx: Context, enum: RuleEnumState) -> None:
        ctx.update({cls.__class__.__name__ + "_state": enum})

    @classmethod
    def check_state(cls, ctx: Context) -> RuleEnumState | None:
        return ctx.get(cls.__class__.__name__ + "_state")

    @classmethod
    def must_be_state(cls, ctx: Context, state: RuleEnumState) -> bool:
        real_state = cls.check_state(ctx)
        if not real_state:
            return False
        return real_state == state

    async def check(self, event: Update, ctx: Context) -> bool:
        if self.check_state(ctx):
            return True

        for enum in self.__enum__:
            ctx_copy = ctx.copy()
            if await check_rule(event.ctx_api, enum.rule, event, ctx_copy):
                ctx.update(ctx_copy)
                self.save_state(ctx, enum)
                return True

        return False


__all__ = ("RuleEnum", "RuleEnumState")



import dataclasses
import types
import typing
from abc import ABC, abstractmethod

from telegrinder.bot.cute_types import BaseCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.modules import logger

T = typing.TypeVar("T")
EventT = typing.TypeVar("EventT", bound=BaseCute, contravariant=True)


def get_union_types(t: types.UnionType) -> tuple[type, ...] | None:
    if type(t) in (types.UnionType, typing._UnionGenericAlias):  # type: ignore
        return tuple(typing.get_origin(x) or x for x in typing.get_args(t))
    return None


def register_manager(return_type: type[typing.Any] | types.UnionType):
    def wrapper(func: typing.Callable[..., typing.Awaitable[typing.Any]]):
        return Manager(get_union_types(return_type) or (return_type,), func)  # type: ignore

    return wrapper


@dataclasses.dataclass(frozen=True)
class Manager:
    types: tuple[type, ...]
    callback: typing.Callable[..., typing.Awaitable]

    async def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        try:
            await self.callback(*args, **kwargs)
        except BaseException as ex:
            logger.exception(ex)


class ABCReturnManager(ABC, typing.Generic[EventT]):
    @abstractmethod
    async def run(self, response: typing.Any, event: EventT, ctx: Context) -> None:
        pass


class BaseReturnManager(ABCReturnManager[EventT]):
    def __repr__(self) -> str:
        return "<{}: {}>".format(
            self.__class__.__name__,
            ", ".join(x.callback.__name__ + "=" + repr(x) for x in self.managers),
        )

    @property
    def managers(self) -> list[Manager]:
        return [
            manager
            for manager in (vars(BaseReturnManager) | vars(self.__class__)).values()
            if isinstance(manager, Manager)
        ]

    @register_manager(Context)
    @staticmethod
    async def ctx_manager(value: Context, event: EventT, ctx: Context) -> None:
        """Basic manager for returning context from handler."""

        ctx.update(value)

    async def run(self, response: typing.Any, event: EventT, ctx: Context) -> None:
        for manager in self.managers:
            if typing.Any in manager.types or any(type(response) is x for x in manager.types):
                await manager(response, event, ctx)

    @typing.overload
    def register_manager(
        self, return_type: type[T]
    ) -> typing.Callable[
        [typing.Callable[[T, EventT, Context], typing.Awaitable[typing.Any]]], Manager
    ]: ...

    @typing.overload
    def register_manager(
        self,
        return_type: tuple[type[T], ...],
    ) -> typing.Callable[
        [typing.Callable[[tuple[T, ...], EventT, Context], typing.Awaitable[typing.Any]]],
        Manager,
    ]: ...

    def register_manager(
        self,
        return_type: type[T] | tuple[type[T], ...],
    ) -> typing.Callable[
        [typing.Callable[[T | tuple[T, ...], EventT, Context], typing.Awaitable[typing.Any]]],
        Manager,
    ]:
        def wrapper(func: typing.Callable[[T, EventT, Context], typing.Awaitable]) -> Manager:
            manager = Manager(get_union_types(return_type) or (return_type,), func)  # type: ignore
            setattr(self.__class__, func.__name__, manager)
            return manager

        return wrapper


__all__ = (
    "ABCReturnManager",
    "BaseReturnManager",
    "Manager",
    "get_union_types",
    "register_manager",
)



import typing

from .base import ComposeError, ScalarNode
from .message import MessageNode


class Text(ScalarNode, str):
    @classmethod
    async def compose(cls, message: MessageNode) -> typing.Self:
        if not message.text:
            raise ComposeError("Message has no text")
        return cls(message.text.unwrap())


__all__ = ("Text",)



from .abc import ABCAPI, Token
from .api import API
from .error import APIError, InvalidTokenError
from .response import APIResponse

__all__ = (
    "ABCAPI",
    "API",
    "APIError",
    "APIResponse",
    "InvalidTokenError",
    "Token",
)



import typing

from telegrinder.bot.dispatch.context import Context
from telegrinder.types.enums import MessageEntityType

from .is_from import IsPrivate
from .markup import Markup, Message
from .message import MessageRule
from .message_entities import MessageEntities


class StartCommand(
    MessageRule,
    requires=[
        IsPrivate()
        & MessageEntities(MessageEntityType.BOT_COMMAND)
        & Markup(["/start <param>", "/start"]),
    ],
):
    def __init__(
        self,
        validator: typing.Callable[[str], typing.Any | None] | None = None,
        *,
        param_required: bool = False,
        alias: str | None = None,
    ) -> None:
        self.param_required = param_required
        self.validator = validator
        self.alias = alias

    async def check(self, _: Message, ctx: Context) -> bool:
        param: str | None = ctx.pop("param", None)
        validated_param = self.validator(param) if self.validator and param is not None else param

        if self.param_required and validated_param is None:
            return False

        ctx.set(self.alias or "param", validated_param)
        return True


__all__ = ("StartCommand",)



import enum
import typing
from reprlib import recursive_repr

from telegrinder.types import Update

T = typing.TypeVar("T")

Key: typing.TypeAlias = str | enum.Enum
AnyValue: typing.TypeAlias = typing.Any


@typing.dataclass_transform(kw_only_default=True, order_default=True)
class Context(dict[str, AnyValue]):
    """Context class for rules and middlewares.
    ```
    class MyRule(ABCRule[T]):
        adapter: ABCAdapter[Update, T] = RawUpdateAdapter()

        async def check(self, event: T, ctx: Context) -> bool:
            ctx.set("value", (await event.ctx_api.get_me()).unwrap())
            return True
    ```
    """

    raw_update: Update

    def __init__(self, **kwargs: AnyValue) -> None:
        cls_vars = vars(self.__class__)
        defaults = {}
        for k in self.__class__.__annotations__:
            if k in cls_vars:
                defaults[k] = cls_vars[k]
                delattr(self.__class__, k)
        dict.__init__(self, **defaults | kwargs)

    @recursive_repr()
    def __repr__(self) -> str:
        return "{}({})".format(
            self.__class__.__name__, ", ".join(f"{k}={v!r}" for k, v in self.items())
        )

    def __setitem__(self, __key: Key, __value: AnyValue) -> None:
        dict.__setitem__(self, self.key_to_str(__key), __value)

    def __getitem__(self, __key: Key) -> AnyValue:
        return dict.__getitem__(self, self.key_to_str(__key))

    def __delitem__(self, __key: Key) -> None:
        dict.__delitem__(self, self.key_to_str(__key))

    def __setattr__(self, __name: str, __value: AnyValue) -> None:
        self.__setitem__(__name, __value)

    def __getattr__(self, __name: str) -> AnyValue:
        return self.__getitem__(__name)

    def __delattr__(self, __name: str) -> None:
        self.__delitem__(__name)

    @staticmethod
    def key_to_str(key: Key) -> str:
        return key if isinstance(key, str) else str(key.value)

    def copy(self) -> typing.Self:
        return self.__class__(**self)

    def set(self, key: Key, value: AnyValue) -> None:
        self[key] = value

    def get(self, key: Key, default: T | None = None) -> T | AnyValue:
        return dict.get(self, key, default)

    def delete(self, key: Key) -> None:
        del self[key]


__all__ = ("Context",)



from .abc import ABCDispatch
from .composition import CompositionDispatch
from .context import Context
from .dispatch import ABCRule, Dispatch, TelegrinderCtx
from .handler import ABCHandler, FuncHandler, MessageReplyHandler
from .middleware import ABCMiddleware
from .process import check_rule, process_inner
from .return_manager import (
    ABCReturnManager,
    BaseReturnManager,
    CallbackQueryReturnManager,
    InlineQueryReturnManager,
    Manager,
    MessageReturnManager,
    register_manager,
)
from .view import (
    ABCStateView,
    ABCView,
    BaseStateView,
    BaseView,
    CallbackQueryView,
    ChatJoinRequestView,
    ChatMemberView,
    InlineQueryView,
    MessageView,
    RawEventView,
    ViewBox,
)
from .waiter_machine import ShortState, WaiterMachine

__all__ = (
    "ABCDispatch",
    "ABCHandler",
    "ABCMiddleware",
    "ABCReturnManager",
    "ABCRule",
    "ABCStateView",
    "ABCView",
    "BaseReturnManager",
    "BaseStateView",
    "BaseView",
    "CallbackQueryReturnManager",
    "CallbackQueryView",
    "ChatJoinRequestView",
    "ChatMemberView",
    "CompositionDispatch",
    "Context",
    "Dispatch",
    "FuncHandler",
    "InlineQueryReturnManager",
    "InlineQueryView",
    "Manager",
    "MessageReplyHandler",
    "MessageReturnManager",
    "MessageView",
    "RawEventView",
    "ShortState",
    "TelegrinderCtx",
    "ViewBox",
    "WaiterMachine",
    "check_rule",
    "process_inner",
    "register_manager",
)



import typing
from abc import ABC, abstractmethod

from fntypes.co import Nothing, Some

from telegrinder.api.abc import ABCAPI
from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.bot.dispatch.handler.abc import ABCHandler
from telegrinder.bot.dispatch.handler.func import FuncHandler
from telegrinder.bot.dispatch.middleware.abc import ABCMiddleware
from telegrinder.bot.dispatch.process import process_inner
from telegrinder.bot.dispatch.return_manager.abc import ABCReturnManager
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.model import Model
from telegrinder.msgspec_utils import Option
from telegrinder.tools.error_handler.error_handler import ABCErrorHandler, ErrorHandler
from telegrinder.types.objects import Update

EventType = typing.TypeVar("EventType", bound=BaseCute)
ErrorHandlerT = typing.TypeVar("ErrorHandlerT", bound=ABCErrorHandler)
MiddlewareT = typing.TypeVar("MiddlewareT", bound=ABCMiddleware)

FuncType: typing.TypeAlias = typing.Callable[
    typing.Concatenate[EventType, ...],
    typing.Coroutine[typing.Any, typing.Any, typing.Any],
]


class ABCView(ABC):
    def __repr__(self) -> str:
        return "<{!r}: {}>".format(
            self.__class__.__name__,
            ", ".join(f"{k}={v!r}" for k, v in self.__dict__.items()),
        )

    @abstractmethod
    async def check(self, event: Update) -> bool:
        pass

    @abstractmethod
    async def process(self, event: Update, api: ABCAPI) -> None:
        pass

    @abstractmethod
    def load(self, external: typing.Self) -> None:
        pass


class ABCStateView(ABCView, typing.Generic[EventType]):
    @abstractmethod
    def get_state_key(self, event: EventType) -> int | None:
        pass


class BaseView(ABCView, typing.Generic[EventType]):
    auto_rules: list[ABCRule[EventType]]
    handlers: list[ABCHandler[EventType]]
    middlewares: list[ABCMiddleware[EventType]]
    return_manager: ABCReturnManager[EventType] | None

    @classmethod
    def get_event_type(cls) -> Option[type[EventType]]:
        for base in cls.__dict__.get("__orig_bases__", ()):
            if issubclass(typing.get_origin(base) or base, ABCView):
                for generic_type in typing.get_args(base):
                    if issubclass(typing.get_origin(generic_type) or generic_type, BaseCute):
                        return Some(generic_type)
        return Nothing()

    @staticmethod
    def get_raw_event(update: Update) -> Option[Model]:
        match update.update_type:
            case Some(update_type):
                return getattr(update, update_type.value)
            case _:
                return Nothing()
    
    @typing.overload
    @classmethod
    def to_handler(
        cls,
        *rules: ABCRule[EventType],
    ) -> typing.Callable[
        [FuncType[EventType]],
        FuncHandler[EventType, FuncType[EventType], ErrorHandler[EventType]],
    ]: ...

    @typing.overload
    @classmethod
    def to_handler(
        cls,
        *rules: ABCRule[EventType],
        error_handler: ErrorHandlerT,
        is_blocking: bool = True,
    ) -> typing.Callable[
        [FuncType[EventType]], FuncHandler[EventType, FuncType[EventType], ErrorHandlerT]
    ]: ...

    @typing.overload
    @classmethod
    def to_handler(
        cls,
        *rules: ABCRule[EventType],
        error_handler: typing.Literal[None] = None,
        is_blocking: bool = True,
    ) -> typing.Callable[
        [FuncType[EventType]],
        FuncHandler[EventType, FuncType[EventType], ErrorHandler[EventType]],
    ]: ...

    @classmethod
    def to_handler(  # type: ignore
        cls,
        *rules: ABCRule[EventType],
        error_handler: ABCErrorHandler | None = None,
        is_blocking: bool = True,
    ):
        def wrapper(func: FuncType[EventType]):
            return FuncHandler(
                func,
                list(rules),
                is_blocking=is_blocking,
                dataclass=None,
                error_handler=error_handler or ErrorHandler(),
            )

        return wrapper

    @typing.overload
    def __call__(
        self,
        *rules: ABCRule[EventType],
    ) -> typing.Callable[
        [FuncType[EventType]],
        FuncHandler[EventType, FuncType[EventType], ErrorHandler[EventType]],
    ]: ...

    @typing.overload
    def __call__(
        self,
        *rules: ABCRule[EventType],
        error_handler: ErrorHandlerT,
        is_blocking: bool = True,
    ) -> typing.Callable[
        [FuncType[EventType]], FuncHandler[EventType, FuncType[EventType], ErrorHandlerT]
    ]: ...

    @typing.overload
    def __call__(
        self,
        *rules: ABCRule[EventType],
        error_handler: typing.Literal[None] = None,
        is_blocking: bool = True,
    ) -> typing.Callable[
        [FuncType[EventType]],
        FuncHandler[EventType, FuncType[EventType], ErrorHandler[EventType]],
    ]: ...

    def __call__(  # type: ignore
        self,
        *rules: ABCRule[EventType],
        error_handler: ABCErrorHandler | None = None,
        is_blocking: bool = True,
    ):
        def wrapper(func: FuncType[EventType]):
            func_handler = FuncHandler(
                func,
                [*self.auto_rules, *rules],
                is_blocking=is_blocking,
                dataclass=None,
                error_handler=error_handler or ErrorHandler(),
            )
            self.handlers.append(func_handler)
            return func_handler

        return wrapper

    def register_middleware(self, *args: typing.Any, **kwargs: typing.Any):
        def wrapper(cls: type[MiddlewareT]) -> type[MiddlewareT]:
            self.middlewares.append(cls(*args, **kwargs))
            return cls

        return wrapper

    async def check(self, event: Update) -> bool:
        match self.get_raw_event(event):
            case Some(e) if issubclass(
                self.get_event_type().expect(
                    "{!r} has no event type in generic.".format(self.__class__.__name__),
                ),
                e.__class__,
            ):
                return True
            case _:
                return False

    async def process(self, event: Update, api: ABCAPI) -> None:
        await process_inner(
            self.get_event_type()
            .unwrap()
            .from_update(
                update=self.get_raw_event(event).unwrap(),
                bound_api=api,
            ),
            event,
            self.middlewares,
            self.handlers,
            self.return_manager,
        )

    def load(self, external: typing.Self) -> None:
        self.auto_rules.extend(external.auto_rules)
        self.handlers.extend(external.handlers)
        self.middlewares.extend(external.middlewares)


class BaseStateView(ABCStateView[EventType], BaseView[EventType], ABC, typing.Generic[EventType]):
    @abstractmethod
    def get_state_key(self, event: EventType) -> int | None:
        pass


__all__ = (
    "ABCStateView",
    "ABCView",
    "BaseStateView",
    "BaseView",
)



from .abc import ABCAdapter
from .errors import AdapterError
from .event import EventAdapter
from .raw_update import RawUpdateAdapter

__all__ = (
    "ABCAdapter",
    "AdapterError",
    "EventAdapter",
    "RawUpdateAdapter",
)



import typing

from fntypes.result import Result

from telegrinder.api import ABCAPI, APIError
from telegrinder.model import get_params
from telegrinder.types import (
    InlineQuery,
    InlineQueryResult,
    InlineQueryResultsButton,
    User,
)

from .base import BaseCute, compose_method_params, shortcut


class InlineQueryCute(BaseCute[InlineQuery], InlineQuery, kw_only=True):
    api: ABCAPI

    @property
    def from_user(self) -> User:
        return self.from_

    @shortcut("answer_inline_query", custom_params={"results"})
    async def answer(
        self,
        results: InlineQueryResult | list[InlineQueryResult],
        inline_query_id: str | None = None,
        cache_time: int | None = None,
        is_personal: bool | None = None,
        next_offset: str | None = None,
        button: InlineQueryResultsButton | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.answer_inline_query()`, see the [documentation](https://core.telegram.org/bots/api#answerinlinequery)

        Use this method to send answers to an inline query. On success, True is returned. 
        No more than 50 results per query are allowed.

        :param inline_query_id: Unique identifier for the answered query.

        :param results: A JSON-serialized array of results for the inline query.

        :param cache_time: The maximum amount of time in seconds that the result of the inline query \
        may be cached on the server. Defaults to 300.

        :param is_personal: Pass True if results may be cached on the server side only for the user that \
        sent the query. By default, results may be returned to any user who sends \
        the same query.

        :param next_offset: Pass the offset that a client should send in the next query with the same text \
        to receive more results. Pass an empty string if there are no more results \
        or if you don't support pagination. Offset length can't exceed 64 bytes. \

        :param button: A JSON-serialized object describing a button to be shown above inline query \
        results.
        """

        params = compose_method_params(
            get_params(locals()),
            self,
            default_params={("inline_query_id", "id")},
        )
        params["results"] = [results] if not isinstance(results, list) else results
        return await self.ctx_api.answer_inline_query(**params)


__all__ = ("InlineQueryCute",)



import typing

from telegrinder.bot.cute_types import MessageCute
from telegrinder.bot.dispatch.return_manager import MessageReturnManager
from telegrinder.types import Update, UpdateType

from .abc import BaseStateView

MessageUpdateType: typing.TypeAlias = typing.Literal[
    UpdateType.MESSAGE,
    UpdateType.BUSINESS_MESSAGE,
    UpdateType.CHANNEL_POST,
    UpdateType.EDITED_BUSINESS_MESSAGE,
    UpdateType.EDITED_CHANNEL_POST,
    UpdateType.EDITED_MESSAGE,
]


class MessageView(BaseStateView[MessageCute]):
    def __init__(self, *, update_type: MessageUpdateType | None = None) -> None:
        self.auto_rules = []
        self.handlers = []
        self.update_type = update_type
        self.middlewares = []
        self.return_manager = MessageReturnManager()

    def get_state_key(self, event: MessageCute) -> int | None:
        return event.chat_id

    async def check(self, event: Update) -> bool:
        if not await super().check(event):
            return False
        return (
            True
            if self.update_type is None
            else self.update_type == event.update_type.unwrap_or_none()
        )


__all__ = ("MessageView",)



import typing

from telegrinder.bot.cute_types import MessageCute

from .base import ComposeError, ScalarNode
from .update import UpdateNode


class MessageNode(ScalarNode, MessageCute):
    @classmethod
    async def compose(cls, update: UpdateNode) -> typing.Self:
        if not update.message:
            raise ComposeError
        return cls(
            **update.message.unwrap().to_dict(),
            api=update.api,
        )


__all__ = ("MessageNode",)



from .html import (
    FormatString,
    HTMLFormatter,
    block_quote,
    bold,
    channel_boost_link,
    code_inline,
    escape,
    invite_chat_link,
    italic,
    link,
    mention,
    pre_code,
    resolve_domain,
    spoiler,
    start_bot_link,
    start_group_link,
    strike,
    tg_emoji,
    underline,
    user_open_message,
)
from .links import (
    get_channel_boost_link,
    get_invite_chat_link,
    get_mention_link,
    get_resolve_domain_link,
    get_start_bot_link,
    get_start_group_link,
    user_open_message_link,
)
from .spec_html_formats import (
    BaseSpecFormat,
    ChannelBoostLink,
    InviteChatLink,
    Link,
    Mention,
    PreCode,
    ResolveDomain,
    SpecialFormat,
    StartBotLink,
    StartGroupLink,
    TgEmoji,
    UserOpenMessage,
)

__all__ = (
    "BaseSpecFormat",
    "ChannelBoostLink",
    "FormatString",
    "HTMLFormatter",
    "InviteChatLink",
    "Link",
    "Mention",
    "PreCode",
    "ResolveDomain",
    "SpecialFormat",
    "StartBotLink",
    "StartGroupLink",
    "TgEmoji",
    "UserOpenMessage",
    "block_quote",
    "bold",
    "channel_boost_link",
    "code_inline",
    "escape",
    "get_channel_boost_link",
    "get_invite_chat_link",
    "get_mention_link",
    "get_resolve_domain_link",
    "get_start_bot_link",
    "get_start_group_link",
    "invite_chat_link",
    "italic",
    "link",
    "mention",
    "pre_code",
    "resolve_domain",
    "spoiler",
    "start_bot_link",
    "start_group_link",
    "strike",
    "tg_emoji",
    "underline",
    "user_open_message",
    "user_open_message_link",
)



import typing

from fntypes.co import Nothing, Some

from telegrinder.api import ABCAPI
from telegrinder.msgspec_utils import Option
from telegrinder.types import Model, Update

from .base import BaseCute

ModelT = typing.TypeVar("ModelT", bound=Model)


class UpdateCute(BaseCute[Update], Update, kw_only=True):
    api: ABCAPI

    @property
    def incoming_update(self) -> Option[Model]:
        return getattr(
            self,
            self.update_type.expect("Update object has no incoming update.").value,
        )

    def get_event(self, event_model: type[ModelT]) -> Option[ModelT]:
        match self.incoming_update:
            case Some(event) if isinstance(event, event_model):
                return Some(event)
        return Nothing()


__all__ = ("UpdateCute",)



from .abc import ABCStateView, ABCView, BaseStateView, BaseView
from .box import ViewBox
from .callback_query import CallbackQueryView
from .chat_join_request import ChatJoinRequestView
from .chat_member import ChatMemberView
from .inline_query import InlineQueryView
from .message import MessageView
from .raw import RawEventView

__all__ = (
    "ABCStateView",
    "ABCView",
    "BaseStateView",
    "BaseView",
    "CallbackQueryView",
    "ChatJoinRequestView",
    "ChatMemberView",
    "InlineQueryView",
    "MessageView",
    "RawEventView",
    "ViewBox",
)



import dataclasses
import typing

from telegrinder.api import API
from telegrinder.types import Chat, Message

from .base import DataNode
from .message import MessageNode


@dataclasses.dataclass
class Source(DataNode):
    api: API
    chat: Chat
    thread_id: int | None = None

    @classmethod
    async def compose(cls, message: MessageNode) -> typing.Self:
        return cls(
            api=message.ctx_api,
            chat=message.chat,
            thread_id=message.message_thread_id.unwrap_or_none(),
        )

    async def send(self, text: str) -> Message:
        result = await self.api.send_message(
            chat_id=self.chat.id,
            message_thread_id=self.thread_id,
            text=text,
        )
        return result.unwrap()


__all__ = ("Source",)



import asyncio
import datetime
import typing

from telegrinder.api.abc import ABCAPI
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.tools.limited_dict import LimitedDict
from telegrinder.types import Update

from .middleware import WaiterMiddleware
from .short_state import Behaviour, EventModel, ShortState

if typing.TYPE_CHECKING:
    from telegrinder.bot.dispatch.view.abc import ABCStateView, BaseStateView

T = typing.TypeVar("T")

Identificator: typing.TypeAlias = str | int
Storage: typing.TypeAlias = dict[str, LimitedDict[Identificator, ShortState[EventModel]]]


class WaiterMachine:
    def __init__(self) -> None:
        self.storage: Storage = {}

    def __repr__(self) -> str:
        return "<{}: storage={!r}>".format(
            self.__class__.__name__,
            self.storage,
        )

    async def drop(
        self,
        state_view: "ABCStateView[EventModel]",
        id: Identificator,
        update: Update,
        **context: typing.Any,
    ) -> None:
        view_name = state_view.__class__.__name__
        if view_name not in self.storage:
            raise LookupError("No record of view {!r} found.".format(view_name))

        short_state = self.storage[view_name].pop(id, None)
        if short_state is None:
            raise LookupError(
                "Waiter with identificator {} is not found for view {!r}".format(
                    id, view_name
                )
            )

        short_state.cancel()
        await self.call_behaviour(
            state_view,
            short_state.event,
            update,
            behaviour=short_state.on_drop_behaviour,
            **context,
        )

    async def wait(
        self,
        state_view: "BaseStateView[EventModel]",
        linked: EventModel | tuple[ABCAPI, Identificator],
        *rules: ABCRule[EventModel],
        default: Behaviour[EventModel] | None = None,
        on_drop: Behaviour[EventModel] | None = None,
        expiration: datetime.timedelta | float | None = None,
    ) -> tuple[EventModel, Context]:
        if isinstance(expiration, int | float):
            expiration = datetime.timedelta(seconds=expiration)

        api: ABCAPI
        key: Identificator
        api, key = (
            linked
            if isinstance(linked, tuple)
            else (linked.ctx_api, state_view.get_state_key(linked))
        )  # type: ignore
        if not key:
            raise RuntimeError("Unable to get state key.")

        view_name = state_view.__class__.__name__
        event = asyncio.Event()
        short_state = ShortState(
            key,
            api,
            event,
            rules,
            expiration=expiration,
            default_behaviour=default,
            on_drop_behaviour=on_drop,
        )
        
        if view_name not in self.storage:
            state_view.middlewares.insert(0, WaiterMiddleware(self, state_view))
            self.storage[view_name] = LimitedDict()

        if (deleted_short_state := self.storage[view_name].set(key, short_state)) is not None:
            deleted_short_state.cancel()
        
        await event.wait()
        self.storage[view_name].pop(key, None)
        return getattr(event, "context")

    async def call_behaviour(
        self,
        view: "ABCStateView[EventModel]",
        event: asyncio.Event | EventModel,
        update: Update,
        behaviour: Behaviour[EventModel] | None = None,
        **context: typing.Any,
    ) -> None:
        # TODO: support param view as a behaviour

        if behaviour is None:
            return

        ctx = Context(**context)
        if isinstance(event, asyncio.Event):
            event, preset_ctx = typing.cast(
                tuple[EventModel, Context],
                getattr(event, "context"),
            )
            ctx.update(preset_ctx)

        if await behaviour.check(event.api, update, ctx):
            await behaviour.run(event, ctx)
        

__all__ = ("WaiterMachine",)



import dataclasses
import enum
import secrets
import typing
from datetime import datetime
from types import NoneType

import msgspec
from fntypes.co import Nothing, Result, Some

from .msgspec_utils import decoder, encoder, get_origin

if typing.TYPE_CHECKING:
    from telegrinder.api.error import APIError

T = typing.TypeVar("T")


MODEL_CONFIG: typing.Final[dict[str, typing.Any]] = {
    "omit_defaults": True,
    "dict": True,
    "rename": {"from_": "from"},
}


@typing.overload
def full_result(
    result: Result[msgspec.Raw, "APIError"],
    full_t: type[T],
) -> Result[T, "APIError"]: ...


@typing.overload
def full_result(
    result: Result[msgspec.Raw, "APIError"],
    full_t: tuple[type[T], ...],
) -> Result[T, "APIError"]: ...


def full_result(
    result: Result[msgspec.Raw, "APIError"],
    full_t: type[T] | tuple[type[T], ...],
) -> Result[T, "APIError"]:
    return result.map(lambda v: decoder.decode(v, type=full_t))  # type: ignore


def get_params(params: dict[str, typing.Any]) -> dict[str, typing.Any]:
    return {
        k: v.unwrap() if isinstance(v, Some) else v
        for k, v in (
            *params.pop("other", {}).items(),
            *params.items(),
        )
        if k != "self" and type(v) not in (NoneType, Nothing)
    }


class Model(msgspec.Struct, **MODEL_CONFIG):
    @classmethod
    def from_bytes(cls, data: bytes) -> typing.Self:
        return decoder.decode(data, type=cls)

    def to_dict(
        self,
        *,
        exclude_fields: set[str] | None = None,
    ) -> dict[str, typing.Any]:
        exclude_fields = exclude_fields or set()
        if "model_as_dict" not in self.__dict__:
            self.__dict__["model_as_dict"] = msgspec.structs.asdict(self)
        return {
            key: value
            for key, value in self.__dict__["model_as_dict"].items()
            if key not in exclude_fields
        }


@dataclasses.dataclass(kw_only=True)
class DataConverter:
    files: dict[str, tuple[str, bytes]] = dataclasses.field(default_factory=lambda: {})

    def __repr__(self) -> str:
        return "<{}: {}>".format(
            self.__class__.__name__,
            ", ".join(f"{k}={v!r}" for k, v in self.converters),
        )

    @property
    def converters(self) -> dict[type[typing.Any], typing.Callable[..., typing.Any]]:
        return {
            get_origin(value.__annotations__["data"]): value
            for key, value in vars(self.__class__).items()
            if key.startswith("convert_") and callable(value)
        }

    @staticmethod
    def convert_enum(data: enum.Enum, _: bool = True) -> typing.Any:
        return data.value

    @staticmethod
    def convert_datetime(data: datetime, _: bool = True) -> int:
        return int(data.timestamp())

    def __call__(self, data: typing.Any, *, serialize: bool = True) -> typing.Any:
        converter = self.get_converter(get_origin(type(data)))
        if converter is not None:
            if isinstance(converter, staticmethod):
                return converter(data, serialize)
            return converter(self, data, serialize)
        return data

    def get_converter(self, t: type[typing.Any]):
        for type, converter in self.converters.items():
            if issubclass(t, type):
                return converter
        return None

    def convert_model(
        self,
        data: Model,
        serialize: bool = True,
    ) -> str | dict[str, typing.Any]:
        converted_dct = self(data.to_dict(), serialize=False)
        return encoder.encode(converted_dct) if serialize is True else converted_dct

    def convert_dct(
        self,
        data: dict[str, typing.Any],
        serialize: bool = True,
    ) -> dict[str, typing.Any]:
        return {
            k: self(v, serialize=serialize)
            for k, v in data.items()
            if type(v) not in (NoneType, Nothing)
        }

    def convert_lst(
        self,
        data: list[typing.Any],
        serialize: bool = True,
    ) -> str | list[typing.Any]:
        converted_lst = [self(x, serialize=False) for x in data]
        return encoder.encode(converted_lst) if serialize is True else converted_lst

    def convert_tpl(
        self,
        data: tuple[typing.Any, ...],
        _: bool = True,
    ) -> str | tuple[typing.Any, ...]:
        if (
            isinstance(data, tuple)
            and len(data) == 2
            and isinstance(data[0], str)
            and isinstance(data[1], bytes)
        ):
            attach_name = secrets.token_urlsafe(16)
            self.files[attach_name] = data
            return "attach://{}".format(attach_name)
        return data


__all__ = (
    "DataConverter",
    "MODEL_CONFIG",
    "Model",
    "full_result",
    "get_params",
)



import dataclasses

import typing_extensions as typing

from telegrinder.api.abc import ABCAPI
from telegrinder.bot.cute_types import BaseCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.process import check_rule
from telegrinder.modules import logger
from telegrinder.tools.error_handler import ABCErrorHandler, ErrorHandler
from telegrinder.types import Update
from telegrinder.types.enums import UpdateType

from .abc import ABCHandler

if typing.TYPE_CHECKING:
    from telegrinder.bot.rules import ABCRule

F = typing.TypeVar(
    "F",
    bound=typing.Callable[typing.Concatenate[typing.Any, ...], typing.Awaitable[typing.Any]],
)
EventT = typing.TypeVar("EventT", bound=BaseCute)
ErrorHandlerT = typing.TypeVar("ErrorHandlerT", bound=ABCErrorHandler, default=ErrorHandler)


@dataclasses.dataclass(repr=False)
class FuncHandler(ABCHandler[EventT], typing.Generic[EventT, F, ErrorHandlerT]):
    func: F
    rules: list["ABCRule[EventT]"]
    _: dataclasses.KW_ONLY
    is_blocking: bool = dataclasses.field(default=True)
    dataclass: type[typing.Any] | None = dataclasses.field(default=dict)
    error_handler: ErrorHandlerT = dataclasses.field(
        default_factory=lambda: typing.cast(ErrorHandlerT, ErrorHandler()),
    )
    preset_context: Context = dataclasses.field(default_factory=lambda: Context())
    update_type: UpdateType | None = dataclasses.field(default=None)

    def __repr__(self) -> str:
        return "<{}: {}={!r} with rules={!r}, dataclass={!r}, error_handler={!r}>".format(
            self.__class__.__name__,
            "blocking function" if self.is_blocking else "function",
            self.func.__name__,
            self.rules,
            self.dataclass,
            self.error_handler,
        )
    
    async def check(self, api: ABCAPI, event: Update, ctx: Context | None = None) -> bool:
        if self.update_type is not None and self.update_type != event.update_type.unwrap_or_none():
            return False
        ctx = ctx or Context()
        temp_ctx = ctx.copy()
        temp_ctx |= self.preset_context

        for rule in self.rules:
            if not await check_rule(api, rule, event, temp_ctx):
                logger.debug("Rule {!r} failed!", rule)
                return False

        ctx |= temp_ctx
        return True

    async def run(self, event: EventT, ctx: Context) -> typing.Any:
        api = event.api
        if self.dataclass is not None:
            if self.update_type is not None:
                update = event.to_dict()[self.update_type.value].unwrap()
                event = (
                    self.dataclass.from_update(update, bound_api=api)  # type: ignore
                    if issubclass(self.dataclass, BaseCute)
                    else self.dataclass(**update.to_dict())
                )
            else:
                event = self.dataclass(**event.to_dict())
        return (await self.error_handler.run(self.func, event, api, ctx)).unwrap()


__all__ = ("FuncHandler",)



import typing
from contextlib import suppress

import msgspec
from fntypes.co import Nothing, Result, Some, Variative, unwrapping

from telegrinder.api import ABCAPI, APIError
from telegrinder.model import get_params
from telegrinder.msgspec_utils import Option, decoder
from telegrinder.types import (
    CallbackQuery,
    InlineKeyboardMarkup,
    InputFile,
    InputMedia,
    LinkPreviewOptions,
    MessageEntity,
    MessageId,
    ReplyParameters,
    User,
)

from .base import BaseCute, compose_method_params, shortcut
from .message import MediaType, MessageCute, ReplyMarkup, execute_method_edit


class CallbackQueryCute(BaseCute[CallbackQuery], CallbackQuery, kw_only=True, dict=True):
    api: ABCAPI

    @property
    def from_user(self) -> User:
        return self.from_

    @property
    def chat_id(self) -> Option[int]:
        """Optional. Message from chat ID. This will be present if the message is sent
        by the bot with the callback button that originated the query."""

        return self.message.map(lambda m: m.v.chat.id)

    @property
    def is_topic_message(self) -> Option[bool]:
        """Optional. True, if the message is a topic message with a name,
        color and icon. This will be present if the message is sent
        by the bot with the callback button that originated the query."""

        return self.message.map(
            lambda m: m.only().map(lambda m: m.is_topic_message.unwrap_or(False)).unwrap_or(False)
        )

    @property
    @unwrapping
    def message_thread_id(self) -> Option[int]:
        """Optional. Unique identifier of the target message thread (for forum supergroups only).
        This will be present if the message is sent
        by the bot with the callback button that originated the query."""

        return (
            self.message.unwrap()
            .only()
            .map(lambda m: m.message_thread_id.unwrap())
            .cast(Some, Nothing)
        )

    @property
    def message_id(self) -> Option[int]:
        """Optional. Unique message identifier inside this chat. This will be present
        if the message is sent by the bot with the callback button that originated the query.
        """

        return self.message.map(lambda m: m.v.message_id)

    def decode_callback_data(self) -> Option[dict[str, typing.Any]]:
        if "cached_callback_data" in self.__dict__:
            return self.__dict__["cached_callback_data"]
        data = Nothing()
        with suppress(msgspec.ValidationError):
            data = Some(decoder.decode(self.data.unwrap()))
        self.__dict__["cached_callback_data"] = data
        return data

    @shortcut("answer_callback_query")
    async def answer(
        self,
        text: str | None = None,
        callback_query_id: str | None = None,
        show_alert: bool | None = None,
        url: str | None = None,
        cache_time: int | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.answer_callback_query()`, see the [documentation](https://core.telegram.org/bots/api#answercallbackquery)

        Use this method to send answers to callback queries sent from inline keyboards. 
        The answer will be displayed to the user as a notification at the top of the 
        chat screen or as an alert. On success, True is returned.

        :param callback_query_id: Unique identifier for the query to be answered.

        :param text: Text of the notification. If not specified, nothing will be shown to the \
        user, 0-200 characters.

        :param show_alert: If True, an alert will be shown by the client instead of a notification at \
        the top of the chat screen. Defaults to false.

        :param url: URL that will be opened by the user's client. If you have created a Game and \
        accepted the conditions via @BotFather, specify the URL that opens your \
        game - note that this will only work if the query comes from a callback_game \
        button. Otherwise, you may use links like t.me/your_bot?start=XXXX that \
        open your bot with a parameter.

        :param cache_time: The maximum amount of time in seconds that the result of the callback query \
        may be cached client-side. Telegram apps will support caching starting \
        in version 3.14. Defaults to 0."""

        params = compose_method_params(
            get_params(locals()), self, default_params={("callback_query_id", "id")}
        )
        return await self.ctx_api.answer_callback_query(**params)

    @shortcut(
        "copy_message",
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def copy(
        self,
        chat_id: int | str | None = None,
        from_chat_id: int | str | None = None,
        message_id: int | None = None,
        message_thread_id: int | None = None,
        caption: str | None = None,
        parse_mode: str | None = None,
        caption_entities: list[MessageEntity] | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageId, APIError]:
        """Shortcut `API.copy_message()`, see the [documentation](https://core.telegram.org/bots/api#copymessage)

        Use this method to copy messages of any kind. Service messages, giveaway 
        messages, giveaway winners messages, and invoice messages can't be copied. 
        A quiz poll can be copied only if the value of the field correct_option_id 
        is known to the bot. The method is analogous to the method forwardMessage, 
        but the copied message doesn't have a link to the original message. Returns 
        the MessageId of the sent message on success.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param from_chat_id: Unique identifier for the chat where the original message was sent (or channel \
        username in the format @channelusername).

        :param message_id: Message identifier in the chat specified in from_chat_id.

        :param caption: New caption for media, 0-1024 characters after entities parsing. If not \
        specified, the original caption is kept.

        :param parse_mode: Mode for parsing entities in the new caption. See formatting options for \
        more details.

        :param caption_entities: A JSON-serialized list of special entities that appear in the new caption, \
        which can be specified instead of parse_mode.

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user.
        """

        return await MessageCute.copy(self, **get_params(locals()))  # type: ignore

    @shortcut("delete_message", custom_params={"message_thread_id"})
    async def delete(
        self,
        chat_id: int | None = None,
        message_id: int | None = None,
        message_thread_id: int | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.delete_message()`, see the [documentation](https://core.telegram.org/bots/api#deletemessage)

        Use this method to delete a message, including service messages, with the 
        following limitations: - A message can only be deleted if it was sent less 
        than 48 hours ago. - Service messages about a supergroup, channel, or forum 
        topic creation can't be deleted. - A dice message in a private chat can only 
        be deleted if it was sent more than 24 hours ago. - Bots can delete outgoing 
        messages in private chats, groups, and supergroups. - Bots can delete incoming 
        messages in private chats. - Bots granted can_post_messages permissions 
        can delete outgoing messages in channels. - If the bot is an administrator 
        of a group, it can delete any message there. - If the bot has can_delete_messages 
        permission in a supergroup or a channel, it can delete any message there. 
        Returns True on success.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_id: Identifier of the message to delete.

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only."""

        return await MessageCute.delete(self, **get_params(locals()))  # type: ignore

    @shortcut(
        "edit_message_text",
        executor=execute_method_edit,
        custom_params={"message_thread_id", "link_preview_options"},
    )
    async def edit_text(
        self,
        text: str,
        inline_message_id: int | None = None,
        chat_id: int | str | None = None,
        message_id: int | None = None,
        message_thread_id: int | None = None,
        parse_mode: str | None = None,
        entities: list[MessageEntity] | None = None,
        link_preview_options: LinkPreviewOptions | dict[str, typing.Any] | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
        **other: typing.Any,
    ) -> Result[Variative[MessageCute, bool], APIError]:
        """Shortcut `API.edit_message_text()`, see the [documentation](https://core.telegram.org/bots/api#editmessagetext)

        Use this method to edit text and game messages. On success, if the edited 
        message is not an inline message, the edited Message is returned, otherwise 
        True is returned.

        :param inline_message_id: Required if chat_id and message_id are not specified. Identifier of the \
        inline message.

        :param chat_id: Required if inline_message_id is not specified. Unique identifier for \
        the target chat or username of the target channel (in the format @channelusername). \

        :param message_id: Required if inline_message_id is not specified. Identifier of the message \
        to edit.

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param text: New text of the message, 1-4096 characters after entities parsing.

        :param parse_mode: Mode for parsing entities in the message text. See formatting options for \
        more details.

        :param entities: A JSON-serialized list of special entities that appear in message text, \
        which can be specified instead of parse_mode.

        :param link_preview_options: Link preview generation options for the message.

        :param reply_markup: A JSON-serialized object for an inline keyboard."""

        ...

    @shortcut(
        "edit_message_live_location",
        executor=execute_method_edit,
        custom_params={"message_thread_id"},
    )
    async def edit_live_location(
        self,
        latitude: float,
        longitude: float,
        inline_message_id: str | None = None,
        message_thread_id: int | None = None,
        chat_id: int | str | None = None,
        message_id: int | None = None,
        horizontal_accuracy: float | None = None,
        heading: int | None = None,
        proximity_alert_radius: int | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
        **other: typing.Any,
    ) -> Result[Variative[MessageCute, bool], APIError]:
        """Shortcut `API.edit_message_live_location()`, see the [documentation](https://core.telegram.org/bots/api#editmessagelivelocation)

        Use this method to edit live location messages. A location can be edited 
        until its live_period expires or editing is explicitly disabled by a call 
        to stopMessageLiveLocation. On success, if the edited message is not an 
        inline message, the edited Message is returned, otherwise True is returned.

        :param inline_message_id: Required if chat_id and message_id are not specified. Identifier of the \
        inline message.

        :param chat_id: Required if inline_message_id is not specified. Unique identifier for \
        the target chat or username of the target channel (in the format @channelusername). \

        :param message_id: Required if inline_message_id is not specified. Identifier of the message \
        to edit.

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param latitude: Latitude of new location.

        :param longitude: Longitude of new location.

        :param horizontal_accuracy: The radius of uncertainty for the location, measured in meters; 0-1500. \

        :param heading: Direction in which the user is moving, in degrees. Must be between 1 and 360 \
        if specified.

        :param proximity_alert_radius: The maximum distance for proximity alerts about approaching another chat \
        member, in meters. Must be between 1 and 100000 if specified.

        :param reply_markup: A JSON-serialized object for a new inline keyboard."""

        ...

    @shortcut(
        "edit_message_caption",
        executor=execute_method_edit,
        custom_params={"message_thread_id"},
    )
    async def edit_caption(
        self,
        caption: str,
        chat_id: int | str | None = None,
        message_id: int | None = None,
        message_thread_id: int | None = None,
        inline_message_id: str | None = None,
        parse_mode: str | None = None,
        caption_entities: list[MessageEntity] | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
        **other: typing.Any,
    ) -> Result[Variative[MessageCute, bool], APIError]:
        """Shortcut `API.edit_message_caption()`, see the [documentation](https://core.telegram.org/bots/api#editmessagecaption)

        Use this method to edit captions of messages. On success, if the edited message 
        is not an inline message, the edited Message is returned, otherwise True 
        is returned.

        :param chat_id: Required if inline_message_id is not specified. Unique identifier for \
        the target chat or username of the target channel (in the format @channelusername). \

        :param message_id: Required if inline_message_id is not specified. Identifier of the message \
        to edit.

        :param inline_message_id: Required if chat_id and message_id are not specified. Identifier of the \
        inline message.

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param caption: New caption of the message, 0-1024 characters after entities parsing. \

        :param parse_mode: Mode for parsing entities in the message caption. See formatting options \
        for more details.

        :param caption_entities: A JSON-serialized list of special entities that appear in the caption, \
        which can be specified instead of parse_mode.

        :param reply_markup: A JSON-serialized object for an inline keyboard."""

        ...

    @shortcut(
        "edit_message_media",
        custom_params={
            "media",
            "type",
            "message_thread_id",
            "caption",
            "parse_mode",
            "caption_entities",
        },
    )
    async def edit_media(
        self,
        media: str | InputFile | InputMedia,
        type: MediaType | None = None,
        caption: str | None = None,
        parse_mode: str | None = None,
        caption_entities: list[MessageEntity] | None = None,
        inline_message_id: str | None = None,
        chat_id: int | str | None = None,
        message_id: int | None = None,
        message_thread_id: int | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
        **other: typing.Any,
    ) -> Result[Variative[MessageCute, bool], APIError]:
        """Shortcut `API.edit_message_media()`, see the [documentation](https://core.telegram.org/bots/api#editmessagemedia)

        Use this method to edit animation, audio, document, photo, or video messages. 
        If a message is part of a message album, then it can be edited only to an audio 
        for audio albums, only to a document for document albums and to a photo or 
        a video otherwise. When an inline message is edited, a new file can't be uploaded; 
        use a previously uploaded file via its file_id or specify a URL. On success, 
        if the edited message is not an inline message, the edited Message is returned, 
        otherwise True is returned.

        :param chat_id: Required if inline_message_id is not specified. Unique identifier for \
        the target chat or username of the target channel (in the format @channelusername). \

        :param message_id: Required if inline_message_id is not specified. Identifier of the message \
        to edit.

        :param inline_message_id: Required if chat_id and message_id are not specified. Identifier of the \
        inline message.

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param media: A JSON-serialized object for a new media content of the message.

        :param caption: Audio caption, 0-1024 characters after entities parsing.

        :param parse_mode: Mode for parsing entities in the audio caption. See formatting options \
        for more details.

        :param caption_entities: A JSON-serialized list of special entities that appear in the caption, \
        which can be specified instead of parse_mode.
        
        :param type: Required if media is not an `str | InputMedia` object. Type of the media, \
        must be one of `photo`, `video`, `animation`, `audio`, `document`.

        :param reply_markup: A JSON-serialized object for a new inline keyboard."""

        return await MessageCute.edit_media(self, **get_params(locals()))  # type: ignore

    @shortcut(
        "edit_message_reply_markup",
        executor=execute_method_edit,
        custom_params={"message_thread_id"},
    )
    async def edit_reply_markup(
        self,
        inline_message_id: str | None = None,
        message_id: int | None = None,
        message_thread_id: int | None = None,
        chat_id: int | str | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
        **other: typing.Any,
    ) -> Result[Variative[MessageCute, bool], APIError]:
        """Shortcut `API.edit_message_reply_markup()`, see the [documentation](https://core.telegram.org/bots/api#editmessagereplymarkup)

        Use this method to edit only the reply markup of messages. On success, if 
        the edited message is not an inline message, the edited Message is returned, 
        otherwise True is returned.

        :param chat_id: Required if inline_message_id is not specified. Unique identifier for \
        the target chat or username of the target channel (in the format @channelusername). \

        :param message_id: Required if inline_message_id is not specified. Identifier of the message \
        to edit.

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param inline_message_id: Required if chat_id and message_id are not specified. Identifier of the \
        inline message.

        :param reply_markup: A JSON-serialized object for an inline keyboard."""

        ...


__all__ = ("CallbackQueryCute",)



import typing
from datetime import datetime

from fntypes.result import Result

from telegrinder.api import ABCAPI, APIError
from telegrinder.model import get_params
from telegrinder.types.objects import ChatMemberUpdated, ChatPermissions, User

from .base import BaseCute, compose_method_params, shortcut
from .utils import compose_chat_permissions


async def chat_member_interaction(
    update: BaseCute[typing.Any],
    method_name: str,
    params: dict[str, typing.Any],
) -> Result[typing.Any, APIError]:
    if isinstance(params.get("permissions"), dict):
        params["permissions"] = compose_chat_permissions(**params["permissions"])
    params = compose_method_params(
        get_params(locals()),
        update,
        default_params={"chat_id", "user_id"},
    )
    return await getattr(update.ctx_api, method_name)(**params)


class ChatMemberShortcuts:
    """Shortcut methods for `ChatMemberUpdatedCute`, `ChatJoinRequestCute` objects."""

    @shortcut("ban_chat_member", executor=chat_member_interaction)
    async def ban_chat_member(
        self,
        chat_id: int | str | None = None,
        user_id: int | None = None,
        until_date: datetime | int | None = None,
        revoke_messages: bool | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.ban_chat_member()`, see the [documentation](https://core.telegram.org/bots/api#banchatmember)

        Use this method to ban a user in a group, a supergroup or a channel. In the case 
        of supergroups and channels, the user will not be able to return to the chat 
        on their own using invite links, etc., unless unbanned first. The bot must 
        be an administrator in the chat for this to work and must have the appropriate 
        administrator rights. Returns True on success.

        :param chat_id: Unique identifier for the target group or username of the target supergroup \
        or channel (in the format @channelusername).

        :param user_id: Unique identifier of the target user.

        :param until_date: Date when the user will be unbanned; Unix time. If user is banned for more \
        than 366 days or less than 30 seconds from the current time they are considered \
        to be banned forever. Applied for supergroups and channels only.

        :param revoke_messages: Pass True to delete all messages from the chat for the user that is being removed. \
        If False, the user will be able to see messages in the group that were sent \
        before the user was removed. Always True for supergroups and channels. \
        """

        ...

    @shortcut("unban_chat_member", executor=chat_member_interaction)
    async def unban_chat_member(
        self,
        chat_id: int | str | None = None,
        user_id: int | None = None,
        only_if_banned: bool | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.unban_chat_member()`, see the [documentation](https://core.telegram.org/bots/api#unbanchatmember)

        Use this method to unban a previously banned user in a supergroup or channel. 
        The user will not return to the group or channel automatically, but will 
        be able to join via link, etc. The bot must be an administrator for this to 
        work. By default, this method guarantees that after the call the user is 
        not a member of the chat, but will be able to join it. So if the user is a member 
        of the chat they will also be removed from the chat. If you don't want this, 
        use the parameter only_if_banned. Returns True on success.

        :param chat_id: Unique identifier for the target group or username of the target supergroup \
        or channel (in the format @channelusername).

        :param user_id: Unique identifier of the target user.

        :param only_if_banned: Do nothing if the user is not banned.
        """

        ...

    @shortcut(
        "restrict_chat_member",
        executor=chat_member_interaction,
        custom_params={"permissions"},
    )
    async def restrict_chat_member(
        self,
        permissions: ChatPermissions | dict[str, typing.Any],
        chat_id: int | str | None = None,
        user_id: int | None = None,
        use_independent_chat_permissions: bool | None = None,
        until_date: datetime | int | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.restrict_chat_member()`, see the [documentation](https://core.telegram.org/bots/api#restrictchatmember)

        Use this method to restrict a user in a supergroup. The bot must be an administrator 
        in the supergroup for this to work and must have the appropriate administrator 
        rights. Pass True for all permissions to lift restrictions from a user. 
        Returns True on success.

        :param chat_id: Unique identifier for the target chat or username of the target supergroup \
        (in the format @supergroupusername).

        :param user_id: Unique identifier of the target user.

        :param permissions: A JSON-serialized object for new user permissions.

        :param use_independent_chat_permissions: Pass True if chat permissions are set independently. Otherwise, the can_send_other_messages \
        and can_add_web_page_previews permissions will imply the can_send_messages, \
        can_send_audios, can_send_documents, can_send_photos, can_send_videos, \
        can_send_video_notes, and can_send_voice_notes permissions; the can_send_polls \
        permission will imply the can_send_messages permission.

        :param until_date: Date when restrictions will be lifted for the user; Unix time. If user is \
        restricted for more than 366 days or less than 30 seconds from the current \
        time, they are considered to be restricted forever.
        """

        ...

    @shortcut("promote_chat_member", executor=chat_member_interaction)
    async def promote_chat_member(
        self,
        chat_id: int | str | None = None,
        user_id: int | None = None,
        is_anonymous: bool | None = None,
        can_manage_chat: bool | None = None,
        can_delete_messages: bool | None = None,
        can_manage_video_chats: bool | None = None,
        can_restrict_members: bool | None = None,
        can_promote_members: bool | None = None,
        can_change_info: bool | None = None,
        can_invite_users: bool | None = None,
        can_post_stories: bool | None = None,
        can_edit_stories: bool | None = None,
        can_delete_stories: bool | None = None,
        can_post_messages: bool | None = None,
        can_edit_messages: bool | None = None,
        can_pin_messages: bool | None = None,
        can_manage_topics: bool | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.promote_chat_member()`, see the [documentation](https://core.telegram.org/bots/api#promotechatmember)

        Use this method to promote or demote a user in a supergroup or a channel. The 
        bot must be an administrator in the chat for this to work and must have the 
        appropriate administrator rights. Pass False for all boolean parameters 
        to demote a user. Returns True on success.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param user_id: Unique identifier of the target user.

        :param is_anonymous: Pass True if the administrator's presence in the chat is hidden.

        :param can_manage_chat: Pass True if the administrator can access the chat event log, get boost list, \
        see hidden supergroup and channel members, report spam messages and ignore \
        slow mode. Implied by any other administrator privilege.

        :param can_delete_messages: Pass True if the administrator can delete messages of other users.

        :param can_manage_video_chats: Pass True if the administrator can manage video chats.

        :param can_restrict_members: Pass True if the administrator can restrict, ban or unban chat members, \
        or access supergroup statistics.

        :param can_promote_members: Pass True if the administrator can add new administrators with a subset \
        of their own privileges or demote administrators that they have promoted, \
        directly or indirectly (promoted by administrators that were appointed \
        by him).

        :param can_change_info: Pass True if the administrator can change chat title, photo and other settings. \

        :param can_invite_users: Pass True if the administrator can invite new users to the chat.

        :param can_post_stories: Pass True if the administrator can post stories to the chat.

        :param can_edit_stories: Pass True if the administrator can edit stories posted by other users.

        :param can_delete_stories: Pass True if the administrator can delete stories posted by other users. \

        :param can_post_messages: Pass True if the administrator can post messages in the channel, or access \
        channel statistics; for channels only.

        :param can_edit_messages: Pass True if the administrator can edit messages of other users and can pin \
        messages; for channels only.

        :param can_pin_messages: Pass True if the administrator can pin messages; for supergroups only. \

        :param can_manage_topics: Pass True if the user is allowed to create, rename, close, and reopen forum \
        topics; for supergroups only.
        """

        ...

    @shortcut("set_chat_administrator_custom_title", executor=chat_member_interaction)
    async def set_chat_administrator_custom_title(
        self,
        custom_title: str,
        chat_id: int | str | None = None,
        user_id: int | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.set_chat_administrator_custom_title()`, see the [documentation](https://core.telegram.org/bots/api#setchatadministratorcustomtitle)

        Use this method to set a custom title for an administrator in a supergroup 
        promoted by the bot. Returns True on success.

        :param chat_id: Unique identifier for the target chat or username of the target supergroup \
        (in the format @supergroupusername).

        :param user_id: Unique identifier of the target user.

        :param custom_title: New custom title for the administrator; 0-16 characters, emoji are not \
        allowed.
        """

        ...


class ChatMemberUpdatedCute(
    BaseCute[ChatMemberUpdated], ChatMemberUpdated, ChatMemberShortcuts, kw_only=True
):
    api: ABCAPI

    @property
    def from_user(self) -> User:
        return self.from_

    @property
    def user_id(self) -> int:
        return self.from_user.id



from .abc import ABCRule, AndRule, NotRule, OrRule
from .callback_data import (
    CallbackDataEq,
    CallbackDataJsonEq,
    CallbackDataJsonModel,
    CallbackDataMap,
    CallbackDataMarkup,
    CallbackQueryDataRule,
    CallbackQueryRule,
    HasData,
)
from .chat_join import (
    ChatJoinRequestRule,
    HasInviteLink,
    InviteLinkByCreator,
    InviteLinkName,
)
from .command import Argument, Command
from .enum_text import EnumTextRule
from .func import FuncRule
from .fuzzy import FuzzyText
from .inline import (
    HasLocation,
    InlineQueryChatType,
    InlineQueryMarkup,
    InlineQueryRule,
    InlineQueryText,
)
from .integer import Integer, IntegerInRange
from .is_from import (
    IsBot,
    IsChat,
    IsChatId,
    IsDiceEmoji,
    IsForum,
    IsForward,
    IsForwardType,
    IsGroup,
    IsLanguageCode,
    IsPremium,
    IsPrivate,
    IsReply,
    IsSuperGroup,
    IsUser,
    IsUserId,
)
from .markup import Markup
from .mention import HasMention
from .message import MessageRule
from .message_entities import HasEntities, MessageEntities
from .regex import Regex
from .rule_enum import RuleEnum
from .start import StartCommand
from .text import HasText, Text, TextMessageRule
from .update import IsUpdate

__all__ = (
    "ABCRule",
    "AndRule",
    "Argument",
    "CallbackDataEq",
    "CallbackDataJsonEq",
    "CallbackDataJsonModel",
    "CallbackDataMap",
    "CallbackDataMarkup",
    "CallbackQueryDataRule",
    "CallbackQueryRule",
    "ChatJoinRequestRule",
    "Command",
    "EnumTextRule",
    "FuncRule",
    "FuzzyText",
    "HasData",
    "HasEntities",
    "HasInviteLink",
    "HasLocation",
    "HasMention",
    "HasText",
    "InlineQueryChatType",
    "InlineQueryMarkup",
    "InlineQueryRule",
    "InlineQueryText",
    "Integer",
    "IntegerInRange",
    "InviteLinkByCreator",
    "InviteLinkName",
    "IsBot",
    "IsChat",
    "IsChatId",
    "IsDiceEmoji",
    "IsForum",
    "IsForward",
    "IsForwardType",
    "IsGroup",
    "IsLanguageCode",
    "IsPremium",
    "IsPrivate",
    "IsReply",
    "IsSuperGroup",
    "IsUpdate",
    "IsUser",
    "IsUserId",
    "Markup",
    "MessageEntities",
    "MessageRule",
    "NotRule",
    "OrRule",
    "Regex",
    "RuleEnum",
    "StartCommand",
    "Text",
    "TextMessageRule",
)



import dataclasses
import typing

from fntypes.option import Nothing

import telegrinder.types
from telegrinder.msgspec_utils import Option

from .base import ComposeError, DataNode, ScalarNode
from .message import MessageNode


@dataclasses.dataclass
class Attachment(DataNode):
    attachment_type: typing.Literal["audio", "document", "photo", "poll", "video"]
    _: dataclasses.KW_ONLY
    audio: Option[telegrinder.types.Audio] = dataclasses.field(default_factory=lambda: Nothing())
    document: Option[telegrinder.types.Document] = dataclasses.field(
        default_factory=lambda: Nothing()
    )
    photo: Option[list[telegrinder.types.PhotoSize]] = dataclasses.field(
        default_factory=lambda: Nothing()
    )
    poll: Option[telegrinder.types.Poll] = dataclasses.field(default_factory=lambda: Nothing())
    video: Option[telegrinder.types.Video] = dataclasses.field(default_factory=lambda: Nothing())

    @classmethod
    async def compose(cls, message: MessageNode) -> "Attachment":
        for attachment_type in ("audio", "document", "photo", "poll", "video"):
            if (attachment := getattr(message, attachment_type, None)) is not None:
                return cls(attachment_type, **{attachment_type: attachment})
        return cls.compose_error("No attachment found in message")


@dataclasses.dataclass
class Photo(DataNode):
    sizes: list[telegrinder.types.PhotoSize]

    @classmethod
    async def compose(cls, attachment: Attachment) -> typing.Self:
        if not attachment.photo:
            raise ComposeError("Attachment is not an photo")
        return cls(attachment.photo.unwrap())


class Video(ScalarNode, telegrinder.types.Video):
    @classmethod
    async def compose(cls, attachment: Attachment) -> telegrinder.types.Video:
        if not attachment.video:
            raise ComposeError("Attachment is not an video")
        return attachment.video.unwrap()


class Audio(ScalarNode, telegrinder.types.Audio):
    @classmethod
    async def compose(cls, attachment: Attachment) -> telegrinder.types.Audio:
        if not attachment.audio:
            raise ComposeError("Attachment is not an audio")
        return attachment.audio.unwrap()


class Document(ScalarNode, telegrinder.types.Document):
    @classmethod
    async def compose(cls, attachment: Attachment) -> telegrinder.types.Document:
        if not attachment.document:
            raise ComposeError("Attachment is not an document")
        return attachment.document.unwrap()


class Poll(ScalarNode, telegrinder.types.Poll):
    @classmethod
    async def compose(cls, attachment: Attachment) -> telegrinder.types.Poll:
        if not attachment.poll:
            raise ComposeError("Attachment is not an poll")
        return attachment.poll.unwrap()


__all__ = (
    "Attachment",
    "Audio",
    "Document",
    "Photo",
    "Poll",
    "Video",
)



import enum
import inspect
import types
import typing

if typing.TYPE_CHECKING:
    from telegrinder.bot.rules.abc import ABCRule

    T = typing.TypeVar("T", bound=ABCRule)

FuncType: typing.TypeAlias = types.FunctionType | typing.Callable[..., typing.Any]
TRANSLATIONS_KEY: typing.Final[str] = "_translations"


def resolve_arg_names(func: FuncType, start_idx: int = 1) -> tuple[str, ...]:
    return func.__code__.co_varnames[start_idx : func.__code__.co_argcount]


def get_default_args(func: FuncType) -> dict[str, typing.Any]:
    fspec = inspect.getfullargspec(func)
    return dict(zip(fspec.args[::-1], (fspec.defaults or ())[::-1]))


def to_str(s: str | enum.Enum) -> str:
    if isinstance(s, enum.Enum):
        return str(s.value)
    return s


def magic_bundle(
    handler: FuncType,
    kw: dict[str | enum.Enum, typing.Any],
    *,
    start_idx: int = 1,
    bundle_ctx: bool = True,
) -> dict[str, typing.Any]:
    names = resolve_arg_names(handler, start_idx=start_idx)
    args = get_default_args(handler)
    args.update({to_str(k): v for k, v in kw.items() if to_str(k) in names})
    if "ctx" in names and bundle_ctx:
        args["ctx"] = kw
    return args


def get_cached_translation(rule: "T", locale: str) -> typing.Optional["T"]:
    translations = getattr(rule, TRANSLATIONS_KEY, {})
    if not translations or locale not in translations:
        return None
    return translations[locale]


def cache_translation(base_rule: "T", locale: str, translated_rule: "T") -> None:
    translations = getattr(base_rule, TRANSLATIONS_KEY, {})
    setattr(base_rule, TRANSLATIONS_KEY, {locale: translated_rule, **translations})


__all__ = (
    "TRANSLATIONS_KEY",
    "cache_translation",
    "get_cached_translation",
    "get_default_args",
    "get_default_args",
    "magic_bundle",
    "resolve_arg_names",
    "to_str",
)



import typing

import fntypes.option
import msgspec
from fntypes.co import Error, Ok, Result, Variative

if typing.TYPE_CHECKING:
    from datetime import datetime

    from fntypes.option import Option
else:
    from datetime import datetime as dt

    Value = typing.TypeVar("Value")

    datetime = type("datetime", (dt,), {})

    class OptionMeta(type):
        def __instancecheck__(cls, __instance: typing.Any) -> bool:
            return isinstance(__instance, fntypes.option.Some | fntypes.option.Nothing)

    class Option(typing.Generic[Value], metaclass=OptionMeta):
        pass


T = typing.TypeVar("T")
Ts = typing.TypeVarTuple("Ts")

DecHook: typing.TypeAlias = typing.Callable[[type[T], typing.Any], object]
EncHook: typing.TypeAlias = typing.Callable[[T], typing.Any]

Nothing: typing.Final[fntypes.option.Nothing] = fntypes.option.Nothing()


def get_origin(t: type[T]) -> type[T]:
    return typing.cast(T, typing.get_origin(t)) or t


def repr_type(t: type) -> str:
    return getattr(t, "__name__", repr(get_origin(t)))


def msgspec_convert(obj: typing.Any, t: type[T]) -> Result[T, msgspec.ValidationError]:
    try:
        return Ok(decoder.convert(obj, type=t, strict=True))
    except msgspec.ValidationError as exc:
        return Error(exc)


def option_dec_hook(tp: type[Option[typing.Any]], obj: typing.Any) -> Option[typing.Any]:
    if obj is None:
        return Nothing
    (value_type,) = typing.get_args(tp) or (typing.Any,)
    return msgspec_convert({"value": obj}, fntypes.option.Some[value_type]).unwrap()


def variative_dec_hook(tp: type[Variative], obj: typing.Any) -> Variative:
    union_types = typing.get_args(tp)

    if isinstance(obj, dict):
        struct_fields_match_sums: dict[type[msgspec.Struct], int] = {
            m: sum(1 for k in obj if k in m.__struct_fields__)
            for m in union_types
            if issubclass(get_origin(m), msgspec.Struct)
        }
        union_types = tuple(t for t in union_types if t not in struct_fields_match_sums)
        reverse = False

        if len(set(struct_fields_match_sums.values())) != len(struct_fields_match_sums.values()):
            struct_fields_match_sums = {
                m: len(m.__struct_fields__) for m in struct_fields_match_sums
            }
            reverse = True

        union_types = (
            *sorted(
                struct_fields_match_sums,
                key=lambda k: struct_fields_match_sums[k],
                reverse=reverse,
            ),
            *union_types,
        )

    for t in union_types:
        match msgspec_convert(obj, t):
            case Ok(value):
                return tp(value)

    raise TypeError(
        "Object of type `{}` does not belong to types `{}`".format(
            repr_type(obj.__class__),
            " | ".join(map(repr_type, union_types)),
        )
    )


class Decoder:
    """Class `Decoder` for `msgspec` module with decode hook
    for objects with the specified type.

    ```
    import enum

    from datetime import datetime as dt

    class Digit(enum.IntEnum):
        ONE = 1
        TWO = 2
        THREE = 3

    decoder = Encoder()
    decoder.dec_hooks[dt] = lambda t, timestamp: t.fromtimestamp(timestamp)

    decoder.dec_hook(dt, 1713354732)  #> datetime.datetime(2024, 4, 17, 14, 52, 12)
    decoder.dec_hook(int, "123")  #> TypeError: Unknown type `int`. You can implement decode hook for this type.

    decoder.convert("123", type=int, strict=False)  #> 123
    decoder.convert(1, type=Digit)  #> <Digit.ONE: 1>

    decoder.decode(b'{"digit":3}', type=dict[str, Digit])  #> {'digit': <Digit.THREE: 3>}
    ```
    """

    def __init__(self) -> None:
        self.dec_hooks: dict[typing.Any, DecHook[typing.Any]] = {
            Option: option_dec_hook,
            Variative: variative_dec_hook,
            datetime: lambda t, obj: t.fromtimestamp(obj),
        }

    def __repr__(self) -> str:
        return "<{}: dec_hooks={!r}>".format(
            self.__class__.__name__,
            self.dec_hooks,
        )

    def add_dec_hook(self, t: T):  # type: ignore
        def decorator(func: DecHook[T]) -> DecHook[T]:
            return self.dec_hooks.setdefault(get_origin(t), func)  # type: ignore

        return decorator

    def dec_hook(self, tp: type[typing.Any], obj: object) -> object:
        origin_type = t if isinstance((t := get_origin(tp)), type) else type(t)
        if origin_type not in self.dec_hooks:
            raise TypeError(
                f"Unknown type `{repr_type(origin_type)}`. "
                "You can implement decode hook for this type."
            )
        return self.dec_hooks[origin_type](tp, obj)

    def convert(
        self,
        obj: object,
        *,
        type: type[T] = dict,
        strict: bool = True,
        from_attributes: bool = False,
        builtin_types: typing.Iterable[type[typing.Any]] | None = None,
        str_keys: bool = False,
    ) -> T:
        return msgspec.convert(
            obj,
            type,
            strict=strict,
            from_attributes=from_attributes,
            dec_hook=self.dec_hook,
            builtin_types=builtin_types,
            str_keys=str_keys,
        )

    @typing.overload
    def decode(self, buf: str | bytes) -> typing.Any: ...

    @typing.overload
    def decode(self, buf: str | bytes, *, strict: bool = True) -> typing.Any: ...

    @typing.overload
    def decode(
        self,
        buf: str | bytes,
        *,
        type: type[T],
        strict: bool = True,
    ) -> T: ...

    def decode(
        self,
        buf: str | bytes,
        *,
        type: type[T] = typing.Any,  # type: ignore
        strict: bool = True,
    ) -> T:
        return msgspec.json.decode(
            buf,
            type=type,
            strict=strict,
            dec_hook=self.dec_hook,
        )


class Encoder:
    """Class `Encoder` for `msgspec` module with encode hooks for objects.

    ```
    from datetime import datetime as dt

    encoder = Encoder()
    encoder.enc_hooks[dt] = lambda d: int(d.timestamp())

    encoder.enc_hook(dt.now())  #> 1713354732
    encoder.enc_hook(123)  #> NotImplementedError: Not implemented encode hook for object of type `int`.

    encoder.encode({'digit': Digit.ONE})  #> '{"digit":1}'
    ```
    """

    def __init__(self) -> None:
        self.enc_hooks: dict[typing.Any, EncHook[typing.Any]] = {
            fntypes.option.Some: lambda opt: opt.value,
            fntypes.option.Nothing: lambda _: None,
            Variative: lambda variative: variative.v,
            datetime: lambda date: int(date.timestamp()),
        }

    def __repr__(self) -> str:
        return "<{}: enc_hooks={!r}>".format(
            self.__class__.__name__,
            self.enc_hooks,
        )

    def add_dec_hook(self, t: type[T]):
        def decorator(func: EncHook[T]) -> EncHook[T]:
            encode_hook = self.enc_hooks.setdefault(get_origin(t), func)
            return func if encode_hook is not func else encode_hook

        return decorator

    def enc_hook(self, obj: object) -> object:
        origin_type = get_origin(obj.__class__)
        if origin_type not in self.enc_hooks:
            raise NotImplementedError(
                "Not implemented encode hook for "
                f"object of type `{repr_type(origin_type)}`."
            )
        return self.enc_hooks[origin_type](obj)

    @typing.overload
    def encode(self, obj: typing.Any) -> str: ...

    @typing.overload
    def encode(self, obj: typing.Any, *, as_str: typing.Literal[True]) -> str: ...

    @typing.overload
    def encode(self, obj: typing.Any, *, as_str: typing.Literal[False]) -> bytes: ...

    def encode(self, obj: typing.Any, *, as_str: bool = True) -> str | bytes:
        buf = msgspec.json.encode(obj, enc_hook=self.enc_hook)
        return buf.decode() if as_str else buf


decoder: typing.Final[Decoder] = Decoder()
encoder: typing.Final[Encoder] = Encoder()


__all__ = (
    "Decoder",
    "Encoder",
    "Nothing",
    "Option",
    "datetime",
    "decoder",
    "encoder",
    "get_origin",
    "msgspec_convert",
    "option_dec_hook",
    "repr_type",
    "variative_dec_hook",
)



import typing
from abc import ABC, abstractmethod


class ABCLoopWrapper(ABC):
    @abstractmethod
    def add_task(self, task: typing.Any) -> None:
        pass

    @abstractmethod
    def run_event_loop(self) -> None:
        pass


__all__ = ("ABCLoopWrapper",)



import asyncio
import dataclasses
import datetime
import typing

from telegrinder.api import ABCAPI
from telegrinder.bot.cute_types import BaseCute
from telegrinder.bot.dispatch.handler.abc import ABCHandler
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.model import Model

if typing.TYPE_CHECKING:
    from .machine import Identificator

T = typing.TypeVar("T", bound=Model)
EventModel = typing.TypeVar("EventModel", bound=BaseCute)

Behaviour: typing.TypeAlias = ABCHandler[T] | None


@dataclasses.dataclass
class ShortState(typing.Generic[EventModel]):
    key: "Identificator"
    ctx_api: ABCAPI
    event: asyncio.Event
    rules: tuple[ABCRule[EventModel], ...]
    _: dataclasses.KW_ONLY
    expiration: dataclasses.InitVar[datetime.timedelta | None] = dataclasses.field(
        default=None,
    )
    default_behaviour: Behaviour[EventModel] | None = dataclasses.field(default=None)
    on_drop_behaviour: Behaviour[EventModel] | None = dataclasses.field(default=None)
    expiration_date: datetime.datetime | None = dataclasses.field(init=False)

    def __post_init__(self, expiration: datetime.timedelta | None = None) -> None:
        self.expiration_date = (datetime.datetime.now() - expiration) if expiration is not None else None
    
    def cancel(self) -> None:
        """Cancel schedule waiters."""

        waiters = typing.cast(
            typing.Iterable[asyncio.Future[typing.Any]],
            self.event._waiters,  # type: ignore
        )
        for future in waiters:
            future.cancel()


__all__ = ("ShortState",)



import typing

from fntypes.result import Error, Ok, Result

from telegrinder.api.abc import ABCAPI
from telegrinder.bot.cute_types import BaseCute
from telegrinder.bot.rules.adapter.abc import ABCAdapter
from telegrinder.bot.rules.adapter.errors import AdapterError
from telegrinder.msgspec_utils import Nothing
from telegrinder.types.objects import Model, Update

CuteT = typing.TypeVar("CuteT", bound=BaseCute)


class EventAdapter(ABCAdapter[Update, CuteT]):
    def __init__(self, event: str | type[Model], cute_model: type[CuteT]) -> None:
        self.event = event
        self.cute_model = cute_model

    def __repr__(self) -> str:
        if isinstance(self.event, str):
            raw_update_type = Update.__annotations__.get(self.event, "Unknown")
            raw_update_type = (
                typing.get_args(raw_update_type)[0].__forward_arg__
                if typing.get_args(raw_update_type)
                else raw_update_type
            )
        else:
            raw_update_type = self.event.__name__
        return "<{}: adapt {} -> {}>".format(
            self.__class__.__name__,
            raw_update_type,
            self.cute_model.__name__,
        )

    async def adapt(self, api: ABCAPI, update: Update) -> Result[CuteT, AdapterError]:
        update_dct = update.to_dict()
        if isinstance(self.event, str):
            if self.event not in update_dct:
                return Error(
                    AdapterError(f"Update is not of event type {self.event!r}."),
                )
            if update_dct[self.event] is Nothing:
                return Error(
                    AdapterError(f"Update is not an {self.event!r}."),
                )
            return Ok(
                self.cute_model.from_update(update_dct[self.event].unwrap(), bound_api=api),
            )
        event = update_dct[update.update_type.unwrap()].unwrap()
        if not update.update_type or not issubclass(event.__class__, self.event):
            return Error(AdapterError(f"Update is not an {self.event.__name__!r}."))
        return Ok(self.cute_model.from_update(event, bound_api=api))


__all__ = ("EventAdapter",)



import abc
import inspect
import typing

ComposeResult: typing.TypeAlias = (
    typing.Coroutine[typing.Any, typing.Any, typing.Any] | typing.AsyncGenerator[typing.Any, None]
)


class ComposeError(BaseException): ...


class Node(abc.ABC):
    node: str = "node"

    @classmethod
    @abc.abstractmethod
    def compose(cls, *args: tuple[typing.Any, ...], **kwargs: typing.Any) -> ComposeResult:
        pass

    @classmethod
    def compose_error(cls, error: str | None = None) -> typing.NoReturn:
        raise ComposeError(error)

    @classmethod
    def get_sub_nodes(cls) -> dict[str, type[typing.Self]]:
        parameters = inspect.signature(cls.compose).parameters

        sub_nodes = {}
        for name, param in parameters.items():
            if param.annotation is inspect._empty:
                continue
            node = param.annotation
            sub_nodes[name] = node
        return sub_nodes

    @classmethod
    def as_node(cls) -> type[typing.Self]:
        return cls

    @classmethod
    def is_generator(cls) -> bool:
        return inspect.isasyncgenfunction(cls.compose)


class DataNode(Node, abc.ABC):
    node = "data"

    @typing.dataclass_transform()
    @classmethod
    @abc.abstractmethod
    async def compose(cls, *args: tuple[typing.Any, ...], **kwargs: typing.Any) -> ComposeResult:
        pass


class ScalarNodeProto(Node, abc.ABC):
    @classmethod
    @abc.abstractmethod
    async def compose(cls, *args: tuple[typing.Any, ...], **kwargs: typing.Any) -> ComposeResult:
        pass


SCALAR_NODE = type("ScalarNode", (), {"node": "scalar"})


if typing.TYPE_CHECKING:

    class ScalarNode(ScalarNodeProto, abc.ABC):
        pass

else:

    def create_node(cls, bases, dct):
        dct.update(cls.__dict__)
        return type(cls.__name__, bases, dct)

    def create_class(name, bases, dct):
        return type(
            "Scalar",
            (SCALAR_NODE,),
            {"as_node": classmethod(lambda cls: create_node(cls, bases, dct))},
        )

    class ScalarNode(ScalarNodeProto, abc.ABC, metaclass=create_class):
        pass


__all__ = (
    "ComposeError",
    "DataNode",
    "Node",
    "SCALAR_NODE",
    "ScalarNode",
)



import dataclasses
import typing

from telegrinder.bot.dispatch.context import Context

from .abc import Message
from .text import TextMessageRule

Validator = typing.Callable[[str], typing.Any | None]


def single_split(s: str, separator: str) -> tuple[str, str]:
    left, *right = s.split(separator, 1)
    return left, (right[0] if right else "")


@dataclasses.dataclass(frozen=True)
class Argument:
    name: str
    validators: list[Validator] = dataclasses.field(default_factory=lambda: [])
    optional: bool = dataclasses.field(default=False, kw_only=True)

    def check(self, data: str) -> typing.Any | None:
        for validator in self.validators:
            data = validator(data)  # type: ignore
            if data is None:
                return None
        return data


class Command(TextMessageRule):
    def __init__(
        self,
        names: str | typing.Iterable[str],
        *arguments: Argument,
        prefixes: tuple[str, ...] = ("/",),
        separator: str = " ",
        lazy: bool = False,
    ) -> None:
        self.names = [names] if isinstance(names, str) else names
        self.arguments = arguments
        self.prefixes = prefixes
        self.separator = separator
        self.lazy = lazy

    def remove_prefix(self, text: str) -> str | None:
        for prefix in self.prefixes:
            if text.startswith(prefix):
                return text.removeprefix(prefix)
        return None

    def parse_argument(
        self,
        arguments: list[Argument],
        data_s: str,
        new_s: str,
        s: str,
    ) -> dict | None:
        argument = arguments[0]
        data = argument.check(data_s)
        if data is None and not argument.optional:
            return None

        if data is None:
            return self.parse_arguments(arguments[1:], s)

        with_argument = self.parse_arguments(arguments[1:], new_s)
        if with_argument is not None:
            return {argument.name: data, **with_argument}

        if not argument.optional:
            return None

        return self.parse_arguments(arguments[1:], s)

    def parse_arguments(self, arguments: list[Argument], s: str) -> dict | None:
        if not arguments:
            return {} if not s else None

        if self.lazy:
            return self.parse_argument(arguments, *single_split(s, self.separator), s)

        all_split = s.split(self.separator)
        for i in range(1, len(all_split) + 1):
            ctx = self.parse_argument(
                arguments,
                self.separator.join(all_split[:i]),
                self.separator.join(all_split[i:]),
                s,
            )
            if ctx is not None:
                return ctx

        return None

    async def check(self, message: Message, ctx: Context) -> bool:
        text = self.remove_prefix(message.text.unwrap())
        if text is None:
            return False

        name, arguments = single_split(text, self.separator)
        if name not in self.names:
            return False

        if not self.arguments:
            return not arguments

        result = self.parse_arguments(list(self.arguments), arguments)
        if result is None:
            return False

        ctx.update(result)
        return True


__all__ = ("Argument", "Command", "single_split")



import inspect
import typing

from telegrinder.bot.dispatch.context import Context
from telegrinder.types import Update

from .abc import ABCAdapter, ABCRule, RawUpdateAdapter, T


class FuncRule(ABCRule, typing.Generic[T]):
    def __init__(
        self,
        func: typing.Callable[[T, Context], typing.Awaitable[bool] | bool],
        adapter: ABCAdapter[Update, T] | None = None,
    ):
        self.func = func
        self.adapter = adapter or RawUpdateAdapter()

    async def check(self, event: T, ctx: Context) -> bool:
        result = self.func(event, ctx)
        if inspect.isawaitable(result):
            return await result
        return result  # type: ignore


__all__ = ("FuncRule",)



import typing
from abc import ABC

from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.bot.dispatch.context import Context

T = typing.TypeVar("T", bound=BaseCute)


class ABCMiddleware(ABC, typing.Generic[T]):
    async def pre(self, event: T, ctx: Context) -> bool: ...

    async def post(self, event: T, responses: list[typing.Any], ctx: Context) -> None: ...


__all__ = ("ABCMiddleware",)



import typing
from collections import UserDict, deque

KT = typing.TypeVar("KT")
VT = typing.TypeVar("VT")


class LimitedDict(UserDict[KT, VT]):
    def __init__(self, *, maxlimit: int = 1000) -> None:
        super().__init__()
        self.maxlimit = maxlimit
        self.queue: deque[KT] = deque(maxlen=maxlimit)

    def set(self, key: KT, value: VT, /) -> VT | None:
        """Set item in the dictionary.
        Returns a value that was deleted when the limit in the dictionary
        was reached, otherwise None.
        """

        deleted_item = None
        if len(self.queue) >= self.maxlimit:
            deleted_item = self.pop(self.queue.popleft(), None)
        if key not in self.queue:
            self.queue.append(key)
        super().__setitem__(key, value)
        return deleted_item

    def __setitem__(self, key: KT, value: VT, /) -> None:
        self.set(key, value)
    
    def __delitem__(self, key: KT) -> None:
        if key in self.queue:
            self.queue.remove(key)
        return super().__delitem__(key)


__all__ = ("LimitedDict",)



import dataclasses

import typing_extensions as typing

from telegrinder.types.enums import UpdateType

from .abc import ABCView
from .callback_query import CallbackQueryView
from .chat_join_request import ChatJoinRequestView
from .chat_member import ChatMemberView
from .inline_query import InlineQueryView
from .message import MessageView
from .raw import RawEventView

CallbackQueryViewT = typing.TypeVar("CallbackQueryViewT", bound=ABCView, default=CallbackQueryView)
ChatJoinRequestViewT = typing.TypeVar(
    "ChatJoinRequestViewT", bound=ABCView, default=ChatJoinRequestView
)
ChatMemberViewT = typing.TypeVar("ChatMemberViewT", bound=ABCView, default=ChatMemberView)
InlineQueryViewT = typing.TypeVar("InlineQueryViewT", bound=ABCView, default=InlineQueryView)
MessageViewT = typing.TypeVar("MessageViewT", bound=ABCView, default=MessageView)
RawEventViewT = typing.TypeVar("RawEventViewT", bound=ABCView, default=RawEventView)


@dataclasses.dataclass(kw_only=True)
class ViewBox(
    typing.Generic[
        CallbackQueryViewT,
        ChatJoinRequestViewT,
        ChatMemberViewT,
        InlineQueryViewT,
        MessageViewT,
        RawEventViewT,
    ],
):
    callback_query: CallbackQueryViewT = dataclasses.field(
        default_factory=lambda: typing.cast(CallbackQueryViewT, CallbackQueryView()),
    )
    chat_join_request: ChatJoinRequestViewT = dataclasses.field(
        default_factory=lambda: typing.cast(ChatJoinRequestViewT, ChatJoinRequestView()),
    )
    chat_member: ChatMemberViewT = dataclasses.field(
        default_factory=lambda: typing.cast(
            ChatMemberViewT, ChatMemberView(update_type=UpdateType.CHAT_MEMBER)
        ),
    )
    my_chat_member: ChatMemberViewT = dataclasses.field(
        default_factory=lambda: typing.cast(
            ChatMemberViewT, ChatMemberView(update_type=UpdateType.MY_CHAT_MEMBER)
        ),
    )
    inline_query: InlineQueryViewT = dataclasses.field(
        default_factory=lambda: typing.cast(InlineQueryViewT, InlineQueryView()),
    )
    message: MessageViewT = dataclasses.field(
        default_factory=lambda: typing.cast(
            MessageViewT, MessageView(update_type=UpdateType.MESSAGE)
        ),
    )
    business_message: MessageViewT = dataclasses.field(
        default_factory=lambda: typing.cast(
            MessageViewT, MessageView(update_type=UpdateType.BUSINESS_MESSAGE)
        ),
    )
    channel_post: MessageViewT = dataclasses.field(
        default_factory=lambda: typing.cast(
            MessageViewT, MessageView(update_type=UpdateType.CHANNEL_POST)
        ),
    )
    edited_message: MessageViewT = dataclasses.field(
        default_factory=lambda: typing.cast(
            MessageViewT, MessageView(update_type=UpdateType.EDITED_MESSAGE)
        ),
    )
    edited_business_message: MessageViewT = dataclasses.field(
        default_factory=lambda: typing.cast(
            MessageViewT,
            MessageView(update_type=UpdateType.EDITED_BUSINESS_MESSAGE),
        ),
    )
    edited_channel_post: MessageViewT = dataclasses.field(
        default_factory=lambda: typing.cast(
            MessageViewT, MessageView(update_type=UpdateType.EDITED_CHANNEL_POST)
        ),
    )
    any_message: MessageViewT = dataclasses.field(
        default_factory=lambda: typing.cast(MessageViewT, MessageView()),
    )
    chat_member_updated: ChatMemberViewT = dataclasses.field(
        default_factory=lambda: typing.cast(ChatMemberViewT, ChatMemberView()),
    )
    raw_event: RawEventViewT = dataclasses.field(
        default_factory=lambda: typing.cast(RawEventViewT, RawEventView()),
    )

    def get_views(self) -> dict[str, ABCView]:
        """Get all views."""

        return {name: view for name, view in self.__dict__.items() if isinstance(view, ABCView)}


__all__ = ("ViewBox",)



from .bot import Telegrinder
from .cute_types import (
    BaseCute,
    CallbackQueryCute,
    ChatJoinRequestCute,
    ChatMemberUpdatedCute,
    InlineQueryCute,
    MessageCute,
    UpdateCute,
)
from .dispatch import (
    ABCDispatch,
    ABCHandler,
    ABCMiddleware,
    ABCReturnManager,
    ABCStateView,
    ABCView,
    BaseReturnManager,
    BaseStateView,
    BaseView,
    CallbackQueryReturnManager,
    CallbackQueryView,
    ChatJoinRequestView,
    ChatMemberView,
    CompositionDispatch,
    Context,
    Dispatch,
    FuncHandler,
    InlineQueryReturnManager,
    Manager,
    MessageReplyHandler,
    MessageReturnManager,
    MessageView,
    RawEventView,
    ShortState,
    ViewBox,
    WaiterMachine,
    register_manager,
)
from .polling import ABCPolling, Polling
from .rules import ABCRule, CallbackQueryRule, MessageRule
from .scenario import ABCScenario, Checkbox, SingleChoice

__all__ = (
    "ABCDispatch",
    "ABCHandler",
    "ABCMiddleware",
    "ABCPolling",
    "ABCReturnManager",
    "ABCRule",
    "ABCScenario",
    "ABCStateView",
    "ABCView",
    "BaseCute",
    "BaseReturnManager",
    "BaseStateView",
    "BaseView",
    "CallbackQueryCute",
    "CallbackQueryReturnManager",
    "CallbackQueryRule",
    "CallbackQueryView",
    "ChatJoinRequestCute",
    "ChatJoinRequestView",
    "ChatMemberUpdatedCute",
    "ChatMemberView",
    "Checkbox",
    "CompositionDispatch",
    "Context",
    "Dispatch",
    "FuncHandler",
    "InlineQueryCute",
    "InlineQueryReturnManager",
    "Manager",
    "MessageCute",
    "MessageReplyHandler",
    "MessageReturnManager",
    "MessageRule",
    "MessageView",
    "Polling",
    "RawEventView",
    "ShortState",
    "SingleChoice",
    "Telegrinder",
    "UpdateCute",
    "ViewBox",
    "WaiterMachine",
    "register_manager",
)



import dataclasses
import typing

from telegrinder.types.enums import ProgrammingLanguage

SpecialFormat: typing.TypeAlias = typing.Union[
    "ChannelBoostLink",
    "InviteChatLink",
    "Link",
    "Mention",
    "PreCode",
    "ResolveDomain",
    "StartBotLink",
    "StartGroupLink",
    "TgEmoji",
    "UserOpenMessage",
]


def is_spec_format(obj: typing.Any) -> typing.TypeGuard[SpecialFormat]:
    return (
        dataclasses.is_dataclass(obj)
        and hasattr(obj, "__formatter_name__")
        and isinstance(obj, BaseSpecFormat)
    )


@dataclasses.dataclass(repr=False)
class BaseSpecFormat:
    __formatter_name__: typing.ClassVar[str] = dataclasses.field(init=False, repr=False)

    def __repr__(self) -> str:
        return f"<Special formatter {self.__class__.__name__!r} -> {self.__formatter_name__!r}>"


@dataclasses.dataclass(repr=False)
class ChannelBoostLink(BaseSpecFormat):
    __formatter_name__ = "channel_boost_link"

    channel_id: str | int
    string: str | None = None


@dataclasses.dataclass(repr=False)
class InviteChatLink(BaseSpecFormat):
    __formatter_name__ = "invite_chat_link"

    invite_link: str
    string: str | None = None


@dataclasses.dataclass(repr=False)
class Mention(BaseSpecFormat):
    __formatter_name__ = "mention"

    string: str
    user_id: int


@dataclasses.dataclass(repr=False)
class Link(BaseSpecFormat):
    __formatter_name__ = "link"

    href: str
    string: str | None = None


@dataclasses.dataclass(repr=False)
class PreCode(BaseSpecFormat):
    __formatter_name__ = "pre_code"

    string: str
    lang: str | ProgrammingLanguage | None = None


@dataclasses.dataclass(repr=False)
class TgEmoji(BaseSpecFormat):
    __formatter_name__ = "tg_emoji"

    string: str
    emoji_id: int


@dataclasses.dataclass(repr=False)
class StartBotLink(BaseSpecFormat):
    __formatter_name__ = "start_bot_link"

    bot_id: str | int
    data: str
    string: str | None


@dataclasses.dataclass(repr=False)
class StartGroupLink(BaseSpecFormat):
    __formatter_name__ = "start_group_link"

    bot_id: str | int
    data: str
    string: str | None = None


@dataclasses.dataclass(repr=False)
class ResolveDomain(BaseSpecFormat):
    __formatter_name__ = "resolve_domain"

    username: str
    string: str | None = None


@dataclasses.dataclass(repr=False)
class UserOpenMessage(BaseSpecFormat):
    __formatter_name__ = "user_open_message"

    user_id: int
    message: str | None = None
    string: str | None = None


__all__ = (
    "BaseSpecFormat",
    "ChannelBoostLink",
    "InviteChatLink",
    "Link",
    "Mention",
    "PreCode",
    "ResolveDomain",
    "SpecialFormat",
    "StartBotLink",
    "StartGroupLink",
    "TgEmoji",
    "UserOpenMessage",
)



from telegrinder.bot.dispatch.context import Context

from .text import Message, TextMessageRule


class Integer(TextMessageRule):
    async def check(self, message: Message, ctx: Context) -> bool:
        return message.text.unwrap().isdigit()


class IntegerInRange(TextMessageRule, requires=[Integer()]):
    def __init__(self, rng: range):
        self.rng = rng

    async def check(self, message: Message, ctx: Context) -> bool:
        return int(message.text.unwrap()) in self.rng


__all__ = ("Integer", "IntegerInRange")



from .base import BaseCute
from .callback_query import CallbackQueryCute
from .chat_join_request import ChatJoinRequestCute
from .chat_member_updated import ChatMemberUpdatedCute
from .inline_query import InlineQueryCute
from .message import MessageCute
from .update import UpdateCute

__all__ = (
    "BaseCute",
    "CallbackQueryCute",
    "ChatJoinRequestCute",
    "ChatMemberUpdatedCute",
    "InlineQueryCute",
    "MessageCute",
    "UpdateCute",
)



import inspect
import typing

from telegrinder.node.base import ComposeError, Node
from telegrinder.node.container import ContainerNode

T = typing.TypeVar("T")


def cast_false_to_none(value: T) -> T | None:
    if value is False:
        return None
    return value


def error_on_none(value: T | None) -> T:
    if value is None:
        raise ComposeError
    return value


def generate(
    subnodes: tuple[type[Node], ...],
    func: typing.Callable[..., typing.Any],
    casts: tuple[typing.Callable, ...] = (cast_false_to_none, error_on_none),
) -> type[ContainerNode]:
    async def compose(**kw: typing.Any) -> typing.Any:
        args = await ContainerNode.compose(**kw)
        result = func(*args)
        if inspect.isawaitable(result):
            result = await result
        for cast in casts:
            result = cast(result)
        return result

    container = ContainerNode.link_nodes(list(subnodes))
    return type("_ContainerNode", (container,), {"compose": compose})


__all__ = ("generate",)



import typing

from telegrinder.bot.cute_types import CallbackQueryCute
from telegrinder.bot.dispatch.waiter_machine import WaiterMachine

from .checkbox import Checkbox

if typing.TYPE_CHECKING:
    from telegrinder.api import API
    from telegrinder.bot.dispatch.view.abc import BaseStateView


class SingleChoice(Checkbox):
    async def handle(self, cb: CallbackQueryCute) -> bool:
        code = cb.data.unwrap().replace(self.random_code + "/", "", 1)
        if code == "ready":
            return False

        for choice in self.choices:
            choice.is_picked = False

        for i, choice in enumerate(self.choices):
            if choice.code == code:
                self.choices[i].is_picked = True
                await cb.ctx_api.edit_message_text(
                    text=self.message,
                    chat_id=cb.message.unwrap().v.chat.id,
                    message_id=cb.message.unwrap().v.message_id,
                    parse_mode=self.PARSE_MODE,
                    reply_markup=self.get_markup(),
                )

        return True

    async def wait(
        self,
        api: "API",
        view: "BaseStateView[CallbackQueryCute]",
    ) -> tuple[str, int]:
        if len(tuple(choice for choice in self.choices if choice.is_picked)) != 1:
            raise ValueError("Exactly one choice must be picked")
        choices, m_id = await super().wait(api, view)
        return tuple(choices.keys())[tuple(choices.values()).index(True)], m_id


__all__ = ("SingleChoice",)



import dataclasses
import typing

from fntypes.result import Error, Ok, Result

from telegrinder.api import ABCAPI
from telegrinder.bot.dispatch.context import Context
from telegrinder.modules import logger
from telegrinder.tools.magic import magic_bundle

from .abc import ABCErrorHandler, EventT, Handler
from .error import CatcherError

F = typing.TypeVar("F", bound="FuncCatcher")
ExceptionT = typing.TypeVar("ExceptionT", bound=BaseException, contravariant=True)
FuncCatcher = typing.Callable[typing.Concatenate[ExceptionT, ...], typing.Awaitable[typing.Any]]


@dataclasses.dataclass(frozen=True, repr=False)
class Catcher(typing.Generic[EventT]):
    func: FuncCatcher[BaseException]
    _: dataclasses.KW_ONLY
    exceptions: list[type[BaseException] | BaseException] = dataclasses.field(
        default_factory=lambda: [],
    )
    logging: bool = dataclasses.field(default=False)
    raise_exception: bool = dataclasses.field(default=False)
    ignore_errors: bool = dataclasses.field(default=False)

    def __repr__(self) -> str:
        return "<Catcher: function={!r}, logging={}, raise_exception={}, ignore_errors={}>".format(
            self.func.__name__,
            self.logging,
            self.raise_exception,
            self.ignore_errors,
        )

    async def __call__(
        self,
        handler: Handler[EventT],
        event: EventT,
        api: ABCAPI,
        ctx: Context,
    ) -> Result[typing.Any, BaseException]:
        try:
            return Ok(await handler(event, **magic_bundle(handler, ctx)))  # type: ignore
        except BaseException as exc:
            return await self.process_exception(api, event, ctx, exc, handler.__name__)

    async def process_exception(
        self,
        api: ABCAPI,
        event: EventT,
        ctx: Context,
        exception: BaseException,
        handler_name: str,
    ) -> Result[typing.Any, BaseException]:
        if self.match_exception(exception):
            logger.debug(
                "Catcher {!r} caught an exception in handler {!r}, " "running catcher...".format(
                    self.func.__name__,
                    handler_name,
                )
            )
            return Ok(
                await self.func(
                    exception,
                    **magic_bundle(self.func, {"event": event, "api": api} | ctx),  # type: ignore
                )
            )
        logger.debug("Failed to match exception {!r}!", exception.__class__.__name__)
        return Error(exception)

    def match_exception(self, exception: BaseException) -> bool:
        for exc in self.exceptions:
            if isinstance(exc, type) and type(exception) is exc:
                return True
            if isinstance(exc, object) and type(exception) is type(exc):
                return True if not exc.args else exc.args == exception.args
        return False


class ErrorHandler(ABCErrorHandler[EventT]):
    def __init__(self, catcher: Catcher[EventT] | None = None, /) -> None:
        self.catcher = catcher

    def __repr__(self) -> str:
        return (
            "<{}: exceptions_handled=[{}], catcher={!r}>".format(
                self.__class__.__name__,
                ", ".join(
                    e.__name__ if isinstance(e, type) else repr(e)
                    for e in self.catcher.exceptions
                ),
                self.catcher,
            )
            if self.catcher is not None
            else "<{}: No catcher>".format(self.__class__.__name__)
        )

    def __call__(
        self,
        *exceptions: type[BaseException] | BaseException,
        logging: bool = False,
        raise_exception: bool = False,
        ignore_errors: bool = False,
    ):
        """Register the catcher.

        :param logging: Logging the result of the catcher at the level 'DEBUG'.
        :param raise_exception: Raise an exception if the catcher hasn't started.
        :param ignore_errors: Ignore errors that may occur.
        """

        def decorator(func: F) -> F:
            if not self.catcher:
                self.catcher = Catcher(
                    func,
                    exceptions=list(exceptions),
                    logging=logging,
                    raise_exception=raise_exception,
                    ignore_errors=ignore_errors,
                )
            return func

        return decorator

    async def process(
        self,
        handler: Handler[EventT],
        event: EventT,
        api: ABCAPI,
        ctx: Context,
    ) -> Result[typing.Any, BaseException]:
        assert self.catcher is not None

        try:
            return await self.catcher(handler, event, api, ctx)
        except BaseException as exc:
            return Error(
                CatcherError(
                    exc,
                    "Exception {} was occurred during the running catcher {!r}.".format(
                        repr(exc), self.catcher.func.__name__
                    ),
                )
            )

    def process_catcher_error(self, error: CatcherError) -> Result[None, BaseException]:
        assert self.catcher is not None

        if self.catcher.raise_exception:
            raise error.exc from None
        if self.catcher.logging:
            logger.error(error.message)
        if not self.catcher.ignore_errors:
            return Error(error.exc)

        return Ok(None)

    async def run(
        self,
        handler: Handler[EventT],
        event: EventT,
        api: ABCAPI,
        ctx: Context,
    ) -> Result[typing.Any, BaseException]:
        if not self.catcher:
            return Ok(await handler(event, **magic_bundle(handler, ctx)))  # type: ignore

        match await self.process(handler, event, api, ctx):
            case Ok(value) as ok:
                if self.catcher.logging:
                    logger.debug(
                        "Catcher {!r} returned a value: {!r}",
                        self.catcher.func.__name__,
                        value,
                    )
                return ok
            case Error(exc) as err:
                if isinstance(exc, CatcherError):
                    return self.process_catcher_error(exc)
                if self.catcher.ignore_errors:
                    return Ok(None)
                return err



import typing
from abc import ABC, abstractmethod

import msgspec

from telegrinder.types import Update


class ABCPolling(ABC):
    offset: int

    @abstractmethod
    async def get_updates(self) -> list[msgspec.Raw]:
        pass

    @abstractmethod
    async def listen(self) -> typing.AsyncGenerator[list[Update], None]:
        yield []

    @abstractmethod
    def stop(self) -> None:
        pass


__all__ = ("ABCPolling",)



import os
import typing

from choicelib import choice_in_order


@typing.runtime_checkable
class JSONModule(typing.Protocol):
    def loads(self, s: str | bytes) -> typing.Any: ...

    def dumps(self, o: typing.Any) -> str: ...


@typing.runtime_checkable
class LoggerModule(typing.Protocol):
    def debug(self, __msg: object, *args: object, **kwargs: object) -> None: ...

    def info(self, __msg: object, *args: object, **kwargs: object) -> None: ...

    def warning(self, __msg: object, *args: object, **kwargs: object) -> None: ...

    def error(self, __msg: object, *args: object, **kwargs: object) -> None: ...

    def critical(self, __msg: object, *args: object, **kwargs: object) -> None: ...

    def exception(self, __msg: object, *args: object, **kwargs: object) -> None: ...

    def set_level(
        self,
        level: typing.Literal[
            "DEBUG",
            "INFO",
            "WARNING",
            "ERROR",
            "CRITICAL",
            "EXCEPTION",
        ],
    ) -> None: ...


logger: LoggerModule
json: JSONModule = choice_in_order(
    ["orjson", "ujson", "hyperjson"], do_import=True, default="telegrinder.msgspec_json"
)
logging_module = choice_in_order(["loguru"], default="logging")
logging_level = os.getenv("LOGGER_LEVEL", default="DEBUG").upper()

if logging_module == "loguru":
    import os
    import sys

    from loguru import logger  # type: ignore

    os.environ.setdefault("LOGURU_AUTOINIT", "0")
    log_format = (
        "<level>{level: <8}</level> | "
        "<lg>{time:YYYY-MM-DD HH:mm:ss}</lg> | "
        "<le>{name}</le>:<le>{function}</le>:"
        "<le>{line}</le> > <lw>{message}</lw>"
    )
    logger.remove()  # type: ignore
    handler_id = logger.add(  # type: ignore
        sink=sys.stderr,
        format=log_format,
        enqueue=True,
        colorize=True,
        level=logging_level,
    )

elif logging_module == "logging":
    """
    This is a workaround for lazy formatting with {} in logging.
    About:
    https://docs.python.org/3/howto/logging-cookbook.html#use-of-alternative-formatting-styles
    """

    import inspect
    import logging
    import sys

    import colorama

    colorama.just_fix_windows_console()  # init & fix console

    FORMAT = (
        "<white>{name: <4} |</white> <level>{levelname: <8}</level>"
        " <white>|</white> <green>{asctime}</green> <white>|</white> <level_module>"
        "{module}</level_module><white>:</white><level_func>"
        "{funcName}</level_func><white>:</white><level_lineno>"
        "{lineno}</level_lineno><white> > </white><level_message>"
        "{message}</level_message>"
    )
    COLORS = {
        "red": colorama.Fore.LIGHTRED_EX,
        "green": colorama.Fore.LIGHTGREEN_EX,
        "blue": colorama.Fore.LIGHTBLUE_EX,
        "white": colorama.Fore.LIGHTWHITE_EX,
        "yellow": colorama.Fore.LIGHTYELLOW_EX,
        "magenta": colorama.Fore.LIGHTMAGENTA_EX,
        "cyan": colorama.Fore.LIGHTCYAN_EX,
        "reset": colorama.Style.RESET_ALL,
    }
    LEVEL_SETTINGS = {
        "INFO": {
            "level": "green",
            "level_module": "blue",
            "level_func": "cyan",
            "level_lineno": "green",
            "level_message": "white",
        },
        "DEBUG": {
            "level": "blue",
            "level_module": "yellow",
            "level_func": "green",
            "level_lineno": "cyan",
            "level_message": "blue",
        },
        "WARNING": {
            "level": "yellow",
            "level_module": "red",
            "level_func": "green",
            "level_lineno": "red",
            "level_message": "yellow",
        },
        "ERROR": {
            "level": "red",
            "level_module": "magenta",
            "level_func": "yellow",
            "level_lineno": "green",
            "level_message": "red",
        },
        "CRITICAL": {
            "level": "cyan",
            "level_module": "yellow",
            "level_func": "yellow",
            "level_lineno": "yellow",
            "level_message": "cyan",
        },
    }
    FORMAT = (
        FORMAT.replace("<white>", COLORS["white"])
        .replace("</white>", COLORS["reset"])
        .replace("<green>", COLORS["green"])
        .replace("</green>", COLORS["reset"])
    )
    LEVEL_FORMATS: dict[str, str] = {}
    for level, settings in LEVEL_SETTINGS.items():
        fmt = FORMAT
        for name, color in settings.items():
            fmt = fmt.replace(f"<{name}>", COLORS[color]).replace(f"</{name}>", COLORS["reset"])
        LEVEL_FORMATS[level] = fmt

    class TelegrinderLoggingFormatter(logging.Formatter):
        def format(self, record: logging.LogRecord) -> str:
            if not record.funcName or record.funcName == "<module>":
                record.funcName = "\b"
            frame = next(
                (
                    frame
                    for frame in inspect.stack()
                    if frame.filename == record.pathname and frame.lineno == record.lineno
                ),
                None,
            )
            if frame:
                module = inspect.getmodule(frame.frame)
                record.module = module.__name__ if module else "<module>"
            return logging.Formatter(
                LEVEL_FORMATS.get(record.levelname),
                datefmt="%Y-%m-%d %H:%M:%S",
                style="{",
            ).format(record)

    class LogMessage:
        def __init__(self, fmt: typing.Any, args: typing.Any, kwargs: typing.Any) -> None:
            self.fmt = fmt
            self.args = args
            self.kwargs = kwargs

        def __str__(self) -> str:
            return self.fmt.format(*self.args, **self.kwargs)

    class TelegrinderLoggingStyleAdapter(logging.LoggerAdapter):
        def __init__(
            self,
            logger: LoggerModule,
            extra: dict[str, typing.Any] | None = None,
        ) -> None:
            super().__init__(logger, extra or {})

        def log(self, level: int, msg: object, *args: object, **kwargs: object) -> None:
            if self.isEnabledFor(level):
                kwargs.setdefault("stacklevel", 2)
                msg, args, kwargs = self.proc(msg, args, kwargs)
                self.logger._log(level, msg, args, **kwargs)

        def proc(
            self,
            msg: object,
            args: tuple[object, ...],
            kwargs: dict[str, object],
        ) -> tuple[LogMessage | object, tuple[object, ...], dict[str, object]]:
            log_kwargs = {
                key: kwargs[key]
                for key in inspect.getfullargspec(self.logger._log).args[1:]
                if key in kwargs
            }

            if isinstance(msg, str):
                msg = LogMessage(msg, args, kwargs)
                args = tuple()
            return msg, args, log_kwargs

    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(TelegrinderLoggingFormatter())
    logger = logging.getLogger("telegrinder")  # type: ignore
    logger.setLevel(logging.getLevelName(logging_level))  # type: ignore
    logger.addHandler(handler)  # type: ignore
    logger = TelegrinderLoggingStyleAdapter(logger)  # type: ignore


def _set_logger_level(level):
    level = level.upper()
    if logging_module == "logging":
        import logging

        logging.getLogger("telegrinder").setLevel(logging.getLevelName(level))
    elif logging_module == "loguru":
        import loguru  # type: ignore

        if handler_id in loguru.logger._core.handlers:  # type: ignore
            loguru.logger._core.handlers[handler_id]._levelno = loguru.logger.level(level).no  # type: ignore


setattr(logger, "set_level", staticmethod(_set_logger_level))  # type: ignore


__all__ = ("json", "logger")



import datetime
import typing

from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.handler.func import FuncHandler
from telegrinder.bot.dispatch.middleware.abc import ABCMiddleware
from telegrinder.bot.dispatch.view.abc import ABCStateView

if typing.TYPE_CHECKING:
    from .machine import WaiterMachine
    from .short_state import ShortState

EventType = typing.TypeVar("EventType", bound=BaseCute)


class WaiterMiddleware(ABCMiddleware[EventType]):
    def __init__(
        self,
        machine: "WaiterMachine",
        view: ABCStateView[EventType],
    ) -> None:
        self.machine = machine
        self.view = view

    async def pre(self, event: EventType, ctx: Context) -> bool:
        if not self.view or not hasattr(self.view, "get_state_key"):
            raise RuntimeError(
                "WaiterMiddleware cannot be used inside a view which doesn't "
                "provide get_state_key (ABCStateView Protocol)."
            )

        view_name = self.view.__class__.__name__
        if view_name not in self.machine.storage:
            return True

        key = self.view.get_state_key(event)
        if key is None:
            raise RuntimeError("Unable to get state key.")

        short_state: "ShortState[EventType] | None" = self.machine.storage[view_name].get(key)
        if not short_state:
            return True
        
        preset_context = Context(short_state=short_state)
        if (
            short_state.expiration_date is not None
            and datetime.datetime.now() >= short_state.expiration_date
        ):
            await self.machine.drop(self.view, short_state.key, ctx.raw_update, **preset_context.copy())
            return True

        handler = FuncHandler(
            self.pass_runtime,
            list(short_state.rules),
            dataclass=None,
            preset_context=preset_context,
        )
        result = await handler.check(event.ctx_api, ctx.raw_update, ctx)

        if result is True:
            await handler.run(event, ctx)

        elif short_state.default_behaviour is not None:
            await self.machine.call_behaviour(
                self.view,
                event,
                ctx.raw_update,
                behaviour=short_state.default_behaviour,
                **handler.preset_context,
            )

        return False

    async def pass_runtime(
        self,
        event: EventType,
        short_state: "ShortState[EventType]",
        ctx: Context,
    ) -> None:
        setattr(short_state.event, "context", (event, ctx))
        short_state.event.set()


__all__ = ("WaiterMiddleware",)



import dataclasses
import typing

from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.process import check_rule
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.node.base import ComposeError, Node
from telegrinder.node.update import UpdateNode


class RuleContext(dict):
    dataclass = dict
    rules: tuple[ABCRule, ...] = ()

    @classmethod
    async def compose(cls, update: UpdateNode):
        ctx = Context()
        for rule in cls.rules:
            if not await check_rule(update.api, rule, update, ctx):
                raise ComposeError
        try:
            return cls.dataclass(**ctx)  # type: ignore
        except Exception as exc:
            raise ComposeError(f"Dataclass validation error: {exc}")

    @classmethod
    def as_node(cls) -> type[typing.Self]:
        return cls

    @classmethod
    def get_sub_nodes(cls) -> dict:
        return {"update": UpdateNode}

    @classmethod
    def is_generator(cls) -> typing.Literal[False]:
        return False

    def __new__(cls, *rules: ABCRule) -> type[Node]:
        return type("_RuleNode", (cls,), {"dataclass": dict, "rules": rules})  # type: ignore

    def __class_getitem__(cls, item: tuple[ABCRule, ...]) -> typing.Self:
        if not isinstance(item, tuple):
            item = (item,)
        return cls(*item)

    @staticmethod
    def generate_dataclass(cls_: type["RuleContext"]):  # noqa: ANN205
        return dataclasses.dataclass(type(cls_.__name__, (object,), dict(cls_.__dict__)))

    def __init_subclass__(cls) -> None:
        if cls.__name__ == "_RuleNode":
            return
        cls.dataclass = cls.generate_dataclass(cls)


__all__ = ("RuleContext",)
